from .auth import AuthHandler
from .client import Client
from .exceptions import *
from .notification import NotificationWebsocket
from .asyncio.client import AsynchronousClient
from .enums import *
from .objects import *
from .util import BeatmapsetSearchFilter


__version__ = '0.4.1'



base_url = "https://osu.ppy.sh/api/v2/"
auth_url = "https://osu.ppy.sh/oauth/authorize/"
token_url = "https://osu.ppy.sh/oauth/token/"


# Info gathered from https://github.com/ppy/osu-web/blob/973315aded8a5762fc00a9f245337802c27bd213/database/mods.json
incompatible_mods = {
    "NoFail": ["SuddenDeath", "Perfect", "AutoPilot", "Relax"],
    "Easy": ["HardRock"],
    "TouchDevice": [],
    "Hidden": [],
    "HardRock": ["Easy", "Mirror"],
    "SuddenDeath": ["NoFail", "Perfect", "Relax", "AutoPilot"],
    "DoubleTime": ["HalfTime"],
    "Relax": ["NoFail", "SuddenDeath", "Perfect", "AutoPilot"],
    "HalfTime": ["DoubleTime", "Nightcore"],
    "Nightcore": ["HalfTime"],
    "Flashlight": [],
    "SpunOut": ["AutoPilot"],
    "AutoPilot": ["NoFail", "SuddenDeath", "Perfect", "Relax", "SpunOut"],
    "Perfect": ["NoFail", "SuddenDeath", "Relax"],
    "FadeIn": ["Hidden", "Flashlight"],
    "Mirror": ["HardRock"],
    **{f"Key{k}": [f"Key{k2}" for k2 in range(4, 10) if k2 != k] for k in range(4, 10)}
}



from ..util import prettify


class Notification:
    """
    Represents a notification object.
    Go to :class:`Details` object to see details for each event

    **Attributes**

    id: :class:`int`

    name: :class:`str`
        Name of the event

    created_at: :class:`str`
        ISO 8601 date

    object_type	: :class:`str`

    object_id: :class:`int`

    source_user_id: :class:`int`

    is_read: :class:`bool`

    details: :class:`Details`
        message_id of last known message (only returned in presence responses)

    **Event Names and Meanings**

    beatmapset_discussion_lock
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who locked discussion

    beatmapset_discussion_post_new
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: Poster of the discussion

    beatmapset_discussion_unlock
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who unlocked discussion

    beatmapset_disqualify
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who disqualified beatmapset

    beatmapset_love
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who promoted beatmapset to Loved

    beatmapset_nominate
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who nominated beatmapset

    beatmapset_qualify
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User whom beatmapset nomination triggered qualification

    beatmapset_remove_from_loved
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who removed beatmapset from Loved

    beatmapset_reset_nominations
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who locked discussion

    channel_message
        object_id: Channel id
        object_type: channel
        source_user_id: User who posted message

    forum_topic_reply
        object_id: Topic id
        object_type: forum_topic
        source_user_id: User who posted message
    """
    __slots__ = (
        "id", "name", "created_at", "object_type", "object_id", "is_read",
        "source_user_id", "details"
    )

    def __init__(self, data):
        self.id = data['id']
        self.name = data['name']
        self.created_at = data['created_at']
        self.object_type = data['object_type']
        self.object_id = data['object_id']
        self.is_read = data['is_read']

        self.source_user_id = data.get('source_user_id', None)
        self.details = Details(data['details'], self.name) if 'details' in data else None

    def __repr__(self):
        return prettify(self, 'name', 'details')


class Details:
    """
    Contains the details for an event

    **Attributes for each Event**

    beatmapset_discussion_lock
        cover_url: :class:`str`
            Beatmap cover
        title: :class:`str`
            Beatmap title
        username: :class:`str`
            Username of source_user_id

    beatmapset_discussion_post_new
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        discussion_id: :class:`int`

        post_i: :class:`int`

        beatmap_id: :class:`int`
            null if posted to general all
        username: :class:`str`
            Username of source_user_id

    beatmapset_discussion_unlock
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        username: :class:`str`
            Username of source_user_id

    beatmapset_disqualify
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        username: :class:`str`
            Username of source_user_id

    beatmapset_love
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        username: :class:`str`
            Username of source_user_id

    beatmapset_nominate
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        username: :class:`str`
            Username of source_user_id

    beatmapset_qualify
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        username: :class:`str`
            Username of source_user_id

    beatmapset_remove_from_loved
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        username: :class:`str`
            Username of source_user_id

    beatmapset_reset_nominations
        title: :class:`str`
            Beatmap title
        cover_url: :class:`str`
            Beatmap cover
        username: :class:`str`
            Username of source_user_id

    channel_message
        title: :class:`str`
            Up to 36 characters of the message (ends with ... when exceeding 36 characters)
        cover_url: :class:`str`
            Avatar of source_user_id
        username: :class:`str`
            Username of source_user_id

    forum_topic_reply
        title: :class:`str`
            Title of the replied topic
        cover_url: :class:`str`
            Topic cover
        post_id: :class:`int`
            Post id
        username: :class:`str`
            Username of source_user_id
    """
    def __init__(self, data, event_name):
        if event_name in ('beatmapset_discussion_lock', 'beatmapset_discussion_unlock',
                          'beatmapset_disqualify', 'beatmapset_love', 'beatmapset_nominate',
                          'beatmapset_qualify', 'beatmapset_remove_from_loved', 'beatmapset_reset_nominations',
                          'channel_message'):
            self.cover_url = data['cover_url']
            self.title = data['title']
            self.username = data['username']
        elif event_name == 'beatmapset_discussion_post_new':
            self.title = data['title']
            self.cover_url = data['cover_url']
            self.discussion_id = data['discussion_id']
            self.post_id = data['post_id']
            self.beatmap_id = data['beatmap_id']
            self.username = data['username']
        elif event_name == 'forum_topic_reply':
            self.title = data['title']
            self.cover_url = data['cover_url']
            self.post_id = data['post_id']
            self.username = data['username']

    def __repr__(self):
        return prettify(self, 'title')



from dateutil import parser

from ..util import prettify


class NewsPost:
    """
    **Attributes**

    author: :class:`str`

    edit_url: :class:`str`
        Link to the file view on GitHub.

    first_image: :class:`str` or :class:`NoneType`
        Link to the first image in the document.

    id: :class:`int`

    published_at: :class:`datetime.datetime`

    slug: :class:`str`
        Filename without the extension, used in URLs.

    title: :class:`str`

    updated_at: :class:`datetime.datetime`

    **Optional Attributes**

    content: :class:`str`
        HTML post content.

    navigation: :class:`Navigation`
        Navigation metadata.

    preview: :class:`str`
        First paragraph of content with HTML markup stripped.
    """
    __slots__ = (
        "author", "edit_url", "first_image", "id",
        "published_at", "slug", "title", "updated_at",
        "content", "navigation", "preview"
    )

    def __init__(self, data):
        self.author = data['author']
        self.edit_url = data['edit_url']
        self.first_image = data['first_image']
        self.id = data['id']
        self.published_at = parser.parse(data['published_at'])
        self.slug = data['slug']
        self.title = data['title']
        self.updated_at = parser.parse(data['updated_at'])
        self.content = data.get("content", "")
        self.navigation = Navigation(data) if "navigation" in data else None
        self.preview = data.get("preview", "")

    def __repr__(self):
        return prettify(self, 'title')


class Navigation:
    """
    **Attributes**

    newer: :class:`NewsPost` or :class:`NoneType`
        null if the next post is not present.

    older: :class:`NewsPost` or :class:`NoneType`
        null if the previous post is not present.
    """
    __slots__ = ("newer", "older")

    def __init__(self, data):
        self.newer = NewsPost(data['newer']) if data["newer"] is not None else None
        self.older = NewsPost(data['older']) if data["older"] is not None else None

    def __repr__(self):
        return prettify(self, 'newer', 'older')



from dateutil import parser
from .user import UserCompact
from ..enums import MatchEventType
from ..util import prettify


class Match:
    __slots__ = ("id", "start_time", "end_time", "name")

    def __init__(self, data):
        self.id = data['id']
        self.name = data['name']
        self.start_time = parser.parse(data['start_time']) if data['start_time'] is not None else None
        self.end_time = parser.parse(data['end_time']) if data['end_time'] is not None else None

    def __repr__(self):
        return prettify(self, 'name', 'start_time')


class MatchExtended(Match):
    """
    Extended version of :class:`Match` that is relevant at the :func:`osu.Client.get_match` endpoint.

    **Attributes**

    events: Sequence[:class:`MatchEvent`]
        List of events that occurred in the match.

    users: Sequence[:class:`UserCompact`]

    first_event_id: :class:`int`

    latest_event_id: :class:`int`

    current_game_id: :class:`int`
    """
    __slots__ = Match.__slots__ + (
        "events", "users", "first_event_id", "latest_event_id", "current_game_id"
    )

    def __init__(self, data):
        super().__init__(data["match"])
        self.events = list(map(MatchEvent, data['events']))
        self.users = list(map(UserCompact, data['users']))
        self.first_event_id = data['first_event_id']
        self.latest_event_id = data['latest_event_id']
        self.current_game_id = data['current_game_id']


class MatchEvent:
    """
    An event that occurred in a match.

    **Attributes**

    id: :class:`int`

    timestamp: :class:`datetime.datetime`

    user_id: :class:`int`

    type: :class:`MatchEventType`

    text: Union[:class:`str`, :class:`NoneType`]
        None unless the event type is :class:`MatchEventType`.OTHER
    """
    __slots__ = ("id", "timestamp", "user_id", "type", "text")

    def __init__(self, data):
        self.id = data['id']
        self.timestamp = parser.parse(data['timestamp'])
        self.user_id = data['user_id']
        self.type = MatchEventType(data['detail']['type'])
        self.text = data['detail']['text'] if 'text' in data['detail'] else None

    def __repr__(self):
        attributes = ('type',) if self.type != MatchEventType.OTHER else ('type', 'text')
        return prettify(self, *attributes)



from dateutil import parser

from ..util import prettify


class KudosuHistory:
    """
    **Attributes**

    id: :class:`int`

    action: :class:`str`
        One of give, vote.give, reset, vote.reset, revoke, or vote.revoke.

    amount: :class:`int`

    model: :class:`str`
        Object type which the exchange happened on (forum_post, etc).

    created_at: :class:`datetime.datetime`

    giver: :class:`Giver` or :class:`NoneType`
        Simple detail of the user who started the exchange.

    post: :class:`Post`
        Simple detail of the object for display.
    """
    __slots__ = (
        "id", "action", "amount", "model", "created_at", "giver", "post"
    )

    def __init__(self, data):
        self.id = data['id']
        self.action = data['action']
        self.amount = data['amount']
        self.model = data['model']
        self.created_at = parser.parse(data['created_at'])
        self.giver = Giver(data['giver']) if data.get('giver') else None
        self.post = Post(data['post'])

    def __repr__(self):
        return prettify(self, 'action', 'amount', 'giver')


class Post:
    """
    **Attributes**

    url: :class:`str` or :class:`NoneType`
        Url of the object.

    title: :class:`str`
        Title of the object. It'll be "[deleted beatmap]" for deleted beatmaps.
    """
    __slots__ = (
        "url", "title"
    )

    def __init__(self, data):
        self.url = data['url']
        self.title = data['title']

    def __repr__(self):
        return prettify(self, 'title')


class Giver:
    """
    **Attributes**

    url: :class:`str`

    username: :class:`str`
    """
    __slots__ = (
        "url", "username"
    )

    def __init__(self, data):
        self.url = data['url']
        self.username = data['username']

    def __repr__(self):
        return prettify(self, 'username')



from .user import CurrentUserAttributes
from .beatmap import BeatmapCompact, BeatmapsetCompact
from dateutil import parser

from ..util import prettify


class BeatmapsetDiscussion:
    """
    Represents a Beatmapset modding discussion

    **Attributes**

    beatmap: :class:`BeatmapCompact` or :class:`NoneType`

    beatmap_id: :class:`int` or :class:`NoneType`

    beatmapset: :class:`BeatmapsetCompact` or :class:`NoneType`

    beatmapset_id: :class:`int`

    can_be_resolved: :class:`bool`

    can_grant_kudosu: :class:`bool`

    created_at: :class:`datetime.datetime`

    current_user_attributes: :class:`CurrentUserAttributes`

    deleted_at: :class:`datetime.datetime` or :class:`NoneType`

    deleted_by_id: :class:`int`

    id: :class:`int`

    kudosu_denied: :class:`bool`

    last_post_at: :class:`datetime.datetime`

    message_type: :class:`str`
        can be any of the following: hype, mapper_note, praise, problem, review, suggestion

    parent_id: :class:`int` or :class:`NoneType`

    posts: :class:`list`
        list contains objects of type :class:`BeatmapsetDiscussionPost`

    resolved: :class:`bool`

    starting_post: :class:`BeatmapsetDiscussionPost`

    timestamp: :class:`int` or :class:`NoneType`

    updated_at: :class:`datetime.datetime`

    user_id: :class:`int`
    """
    __slots__ = (
        "beatmap", "beatmap_id", "beatmapset", "beatmapset_id", "can_be_resolved", "can_grant_kudosu",
        "created_at", "current_user_attributes", "deleted_at", "deleted_by_id", "id", "kudosu_denied",
        "last_post_at", "message_type", "parent_id", "posts", "resolved", "starting_post", "timestamp",
        "updated_at", "user_id", "votes"
    )

    def __init__(self, data):
        self.beatmap = BeatmapCompact(data['beatmap']) if data.get('beatmap') is not None else None
        self.beatmap_id = data['beatmap_id']
        self.beatmapset = BeatmapsetCompact(data['beatmapset']) if data.get('beatmapset') is not None else None
        self.beatmapset_id = data['beatmapset_id']
        self.can_be_resolved = data['can_be_resolved']
        self.can_grant_kudosu = data['can_grant_kudosu']
        self.created_at = parser.parse(data['created_at'])
        self.current_user_attributes = CurrentUserAttributes(data['current_user_attributes'],
                                                             'BeatmapsetDiscussionPermissions') \
            if 'current_user_attributes' in data else None
        self.deleted_at = parser.parse(data['deleted_at']) if data['deleted_at'] is not None else None
        self.deleted_by_id = data['deleted_by_id'] if 'deleted_by_id' in data else None
        self.id = data['id']
        self.kudosu_denied = data['kudosu_denied']
        self.last_post_at = parser.parse(data['last_post_at'])
        self.message_type = data['message_type']
        self.parent_id = data['parent_id'] if 'parent_id' in data else None
        self.posts = list(map(BeatmapsetDiscussionPost, data['posts'])) if 'posts' in data else None
        self.resolved = data['resolved']
        self.starting_post = BeatmapsetDiscussionPost(data['starting_post']) if 'starting_post' in data else None
        self.timestamp = data['timestamp'] if 'timestamp' in data else None
        self.updated_at = parser.parse(data['updated_at'])
        self.user_id = data['user_id']

    def __repr__(self):
        return prettify(self, 'beatmapset', 'starting_post')


class BeatmapsetDiscussionPost:
    """
    Represents a post in a :class:`BeatmapsetDiscussion`.

    **Attributes**

    beatmapset_discussion_id: :class:`int`

    created_at: :class:`datetime.datetime`

    deleted_at: :class:`datetime.datetime` or :class:`NoneType`

    deleted_by_id: :class:`int` or :class:`NoneType`

    id: :class:`int`

    last_editor_id: :class:`int` or :class:`NoneType`

    message: :class:`str`

    system: :class:`bool`

    updated_at: :class:`datetime.datetime`

    user_id: :class:`int`
    """
    __slots__ = (
        "beatmapset_discussion_id", "created_at", "deleted_at", "deleted_by_id", "id",
        "last_editor_id", "message", "system", "updated_at", "user_id"
    )

    def __init__(self, data):
        self.beatmapset_discussion_id = data['beatmapset_discussion_id']
        self.created_at = parser.parse(data['created_at'])
        self.deleted_at = parser.parse(data['deleted_at']) if data['deleted_at'] is not None else None
        self.deleted_by_id = data['deleted_by_id']
        self.id = data['id']
        self.last_editor_id = data['last_editor_id']
        self.message = data['message']
        self.system = data['system']
        self.updated_at = parser.parse(data['updated_at'])
        self.user_id = data['user_id']

    def __repr__(self):
        return prettify(self, 'user_id', 'message', 'beatmapset_discussion_id')


class BeatmapsetDiscussionVote:
    """
    Represents a vote on a :class:`BeatmapsetDiscussion`.

    **Attributes**

    beatmapset_discussion_id: :class:`int`

    created_at: :class:`datetime.datetime`

    id: :class:`int`

    score: :class:`int`

    updated_at: :class:`datetime.datetime`

    user_id: :class:`int`
    """
    __slots__ = (
        "beatmapset_discussion_id", "created_at", "id", "score",
        "updated_at", "user_id"
    )

    def __init__(self, data):
        self.beatmapset_discussion_id = data['beatmapset_discussion_id']
        self.created_at = parser.parse(data['created_at'])
        self.id = data['id']
        self.score = data['score']
        self.updated_at = parser.parse(data['updated_at'])
        self.user_id = data['user_id']

    def __repr__(self):
        return prettify(self, 'beatmapset_discussion_id', 'score', 'user_id')



from .enums import Mods, Enum, BeatmapsetSearchSort, BeatmapsetSearchGeneral, BeatmapsetSearchPlayed, \
    BeatmapsetSearchStatus, BeatmapsetSearchExtra, ScoreRank, GameModeInt
from typing import Sequence, Union


def check_scope(func):
    def check(self, other):
        if not isinstance(other, self.__class__) and type(other) != str:
            raise TypeError(f"Cannot compare type Scope with type {type(other).__name__}")
        if isinstance(other, self.__class__):
            other = other.scope
        if other not in self.valid_scopes:
            raise NameError(f"{other} is not a valid scope.")
        return func(self, other)
    return check


def create_autoclass_for_sphinx():
    with open('objects.py', 'r') as f:
        info = f.readline()
        while info:
            if info.startswith('class'):
                name = info.split()[1]
                if '(' in name:
                    name = name.split('(')[0]
                else:
                    name = name[:-1]
                content = f"{name}\n{'^' * len(name)}\n.. autoclass:: osu.{name}\n   :members:\n\n"
                with open('text.txt', 'a') as f2:
                    f2.write(content)
                    f2.close()
            info = f.readline()
        f.close()


def parse_mods_arg(mods):
    if mods is None:
        return
    if isinstance(mods, Mods):
        return mods.value
    if isinstance(mods, Sequence):
        if len(mods) == 0:
            return
        return Mods.parse_any_list(mods).value
    raise TypeError(f"mods argument must be of type Mods or Sequence, not {type(mods)}")


def prettify(cls: object, *fields: str) -> str:
    d = {s: getattr(cls, s) for s in fields}
    return cls.__class__.__qualname__ + '(' + ', '.join([f"{k}={d[k]!r}" for k in d]) + ')'


def parse_enum_args(*args):
    args = [arg.value if isinstance(arg, Enum) else arg for arg in args]
    return args if len(args) != 1 else args[0]


class Util:
    @staticmethod
    def int(value):
        if value is None:
            return
        return int(value)

    @staticmethod
    def float(value):
        if value is None:
            return
        return float(value)


def returns_self(func):
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        return self
    return wrapper


class BeatmapsetSearchFilter:
    """
    Util class that helps for filtering in :func:`Client.search_beatmapsets`.

    Read about each filter under its corresponding function.

    All set functions return the instance of the class to allow for chaining.
    """
    def __init__(self):
        self._filters = {
            "query": None,
            "sort": None,
            "nsfw": None,
            "played": None,
            "r": None,
            "m": None,
            "l": None,
            "g": None,
            "e": None,
            "c": None,
            "s": None,
        }

    def set_query(self, query: str) -> 'BeatmapsetSearchFilter':
        """
        Set the query to search for.

        query: :class:`str`
        """
        self._filters["query"] = query
        return self

    def set_sort(self, sort: Union[BeatmapsetSearchSort, str]) -> 'BeatmapsetSearchFilter':
        """
        Set the sort order of the search.

        sort: Union[:class:`BeatmapsetSearchSort`, :class:`str`]
        """
        self._filters["sort"] = sort.value if isinstance(sort, BeatmapsetSearchSort) else sort
        return self

    def set_nsfw(self, nsfw: bool) -> 'BeatmapsetSearchFilter':
        """
        Set whether to include NSFW in the search.

        nsfw: :class:`bool`
        """
        self._filters["nsfw"] = nsfw
        return self

    def set_played(self, played: Union[BeatmapsetSearchPlayed, str]) -> 'BeatmapsetSearchFilter':
        """
        Set whether to include played and unplayed beatmapsets in the search.

        played: Union[:class:`BeatmapsetSearchPlayed`, :class:`str`]
        """
        self._filters["played"] = played.value if isinstance(played, BeatmapsetSearchPlayed) else played
        return self

    def set_ranked(self, rank: Union[ScoreRank, str]) -> 'BeatmapsetSearchFilter':
        """
        Filter by rank achieved.

        rank: Union[:class:`ScoreRank`, :class:`str`]
        """
        self._filters["r"] = rank.value if isinstance(rank, ScoreRank) else rank
        return self

    def set_mode(self, mode: Union[GameModeInt, str]) -> 'BeatmapsetSearchFilter':
        """
        Set the game mode to filter by.

        mode: Union[:class:`GameModeStr`, :class:`str`]
        """
        self._filters["m"] = mode.value if isinstance(mode, GameModeInt) else mode
        return self

    def set_language(self, language: str) -> 'BeatmapsetSearchFilter':
        """
        Set the language to filter by.

        language: :class:`str`
        """
        self._filters["l"] = language
        return self

    def set_genre(self, genre: str) -> 'BeatmapsetSearchFilter':
        """
        Set the genre to filter by.

        genre: :class:`str`
        """
        self._filters["g"] = genre
        return self

    def set_extra(self, extras: Sequence[Union[BeatmapsetSearchExtra, str]]) -> 'BeatmapsetSearchFilter':
        """
        Set the extras to filter by.

        extras: Sequence[Union[:class:`BeatmapsetSearchExtra`, :class:`str`]]
        """
        self._filters["e"] = ".".join(list(map(
            lambda x: x.value if isinstance(x, BeatmapsetSearchExtra) else x, extras)))
        return self

    def set_generals(self, generals: Sequence[Union[BeatmapsetSearchGeneral, str]]) -> 'BeatmapsetSearchFilter':
        """
        Set the generals to filter by.

        generals: Sequence[Union[:class:`BeatmapsetSearchGeneral`, :class:`str`]]
        """
        self._filters["c"] = ".".join(list(map(
            lambda x: x.value if isinstance(x, BeatmapsetSearchGeneral) else x, generals)))
        return self

    def set_status(self, status: Union[BeatmapsetSearchStatus, str]) -> 'BeatmapsetSearchFilter':
        """
        Set the status to filter by.

        status: Union[:class:`BeatmapsetSearchStatus`, :class:`str`]
        """
        self._filters["s"] = status.value if isinstance(status, BeatmapsetSearchStatus) else status
        return self

    @property
    def filters(self):
        """
        Dictionary of all filters only including ones that have been set.
        """
        return {k: v for k, v in self._filters.items() if v is not None}



import time
import asyncio
import aiohttp

from ..exceptions import ScopeException
from ..constants import base_url


class AsynchronousHTTPHandler:
    def __init__(self, client, request_wait_time, limit_per_minute):
        self.client = client
        self.rate_limit = RateLimitHandler(request_wait_time, limit_per_minute)

    def get_headers(self, requires_auth=True, **kwargs):
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            **{str(key): str(value) for key, value in kwargs.items() if value is not None}
        }
        if requires_auth:
            headers['Authorization'] = f"Bearer {self.client.auth.token}"
        return headers

    async def make_request(self, method, path, data=None, headers=None, is_download=False, **kwargs):
        if headers is None:
            headers = {}
        if data is None:
            data = {}

        if path.requires_auth and self.client.auth is None:
            raise ScopeException("You need to be authenticated to make this request.")

        if path.requires_auth and path.scope not in self.client.auth.scope:
            raise ScopeException(f"You don't have the {path.scope} scope, which is required to make this request.")

        if path.requires_user and not self.client.auth.has_user:
            raise ScopeException("This request requires a user. You need either a delegate scope or "
                                 "to register OAuth with Authorization Code Grant.")

        headers = self.get_headers(path.requires_auth, **headers)
        params = {str(key): value for key, value in kwargs.items() if value is not None}

        async with aiohttp.ClientSession() as session:
            if not self.rate_limit.can_request:
                await self.rate_limit.wait()

            self.rate_limit.request_used()
            async with session.request(method, base_url + path.path, headers=headers, data=data, params=params) as resp:
                resp.raise_for_status()
                return await resp.json() if not is_download else resp.content


class RateLimitHandler:
    def __init__(self, request_wait_limit, limit_per_minute):
        self.wait_limit = request_wait_limit
        self.limit = limit_per_minute
        self.requests = []
        self.waiting = False

    def request_used(self):
        self.requests.append(time.perf_counter())
        self.reset()

    async def wait(self):
        while self.waiting:
            await asyncio.sleep(0.1)
        self.waiting = True
        if self.can_request:  # Check again due to asynchronous running
            self.waiting = False
            return
        next_available_request = self.wait_limit-(time.perf_counter()-self.last_request)
        if len(self.requests) >= self.limit:
            next_available_request = max(next_available_request, self.requests[0]+60-time.perf_counter())
        await asyncio.sleep(next_available_request)
        self.waiting = False

    def reset(self):
        while len(self.requests) > 0:
            if self.requests[0] + 60 < time.perf_counter():
                self.requests.pop(0)
            else:
                break

    @property
    def can_request(self):
        self.reset()
        return time.perf_counter()-self.last_request >= self.wait_limit and len(self.requests) < self.limit

    @property
    def last_request(self):
        return self.requests[-1] if len(self.requests) > 0 else 0



from dateutil import parser

from ..util import prettify


class Build:
    """
    **Attributes**

    created_at: :class:`datetime.datetime`

    display_version: :class:`str`

    id: :class:`int`

    update_stream: :class:`UpdateStream` or :class:`NoneType`

    users: :class:`int`

    version :class:`str` or :class:`NoneType`

    **Optional Attributes**

    changelog_entries: :class:`list`
        list of :class:`ChangelogEntry` objects. If the build has no changelog entries, a placeholder is generated.

    versions: :class:`Versions`
    """
    __slots__ = (
        "created_at", "display_version", "id", "update_stream", "users", "version",
        "changelog_entries", "versions"
    )

    def __init__(self, data):
        self.created_at = parser.parse(data["created_at"])
        self.display_version = data['display_version']
        self.id = data['id']
        self.update_stream = UpdateStream(data['update_stream']) if data.get('update_stream') is not None else None
        self.users = data['users']
        self.version = data['version']
        self.changelog_entries = list(map(ChangelogEntry, data['changelog_entries'])) \
            if data.get("changelog_entries") is not None else []
        self.versions = Versions(data) if "versions" in data else None

    def __repr__(self):
        return prettify(self, 'id', 'version')


class Versions:
    """
    **Optional Attributes**

    next: :class:`Build` or :class:`NoneType`
        May be null if there is not a next build.

    previous: :class:`Build` or :class:`NoneType`
        May be null if there is not a previous build.
    """
    __slots__ = ("next", "previous")

    def __init__(self, data):
        self.next = Build(data['next']) if data.get('next') is not None else None
        self.previous = Build(data['previous']) if data.get('previous') is not None else None

    def __repr__(self):
        fields = [getattr(self, slot) for slot in self.__slots__ if getattr(self, slot, None)]
        return prettify(self, *fields)


class UpdateStream:
    """
    **Attributes**

    display_name: :class:`str` or :class:`NoneType`

    id: :class:`int`

    is_featured: :class:`bool`

    name: :class:`str`

    **Optional Attributes**

    latest_build: :class:`Build`

    user_count: :class:`int`
    """
    __slots__ = (
        "display_name", "id", "is_featured", "name",
        "latest_build", "user_count"
    )

    def __init__(self, data):
        self.display_name = data['display_name']
        self.id = data['id']
        self.is_featured = data['is_featured']
        self.name = data['name']
        self.latest_build = Build(data['latest_build']) if 'latest_build' in data else None
        self.user_count = data.get('user_count', None)

    def __repr__(self):
        return prettify(self, 'display_name')


class ChangelogEntry:
    """
    **Attributes**

    category: :class:`str`

    created_at: :class:`datetime.datetime` or :class:`NoneType`

    github_pull_request_id: :class:`int` or :class:`NoneType`

    github_url: :class:`str` or :class:`NoneType`

    id: :class:`int` or :class:`NoneType`

    major: :class:`bool`

    repository: :class:`str` or :class:`NoneType`

    title: :class:`str` or :class:`NoneType`

    type: :class:`str`

    url: :class:`str` or :class:`NoneType`

    **Optional Attributes**

    github_user: :class:`GithubUser`
        If the changelog entry has no GitHub user, a placeholder is generated.

    message: :class:`str`
        Entry message in Markdown format. Embedded HTML is allowed.

    message_html: :class:`str`
        Entry message in HTML format.
    """
    __slots__ = (
        "category", "created_at", "github_pull_request_id", "github_url",
        "id", "major", "repository", "title", "type", "url", "github_user",
        "message", "message_html"
    )

    def __init__(self, data):
        self.category = data['category']
        self.created_at = parser.parse(data['created_at']) if data['created_at'] is not None else None
        self.github_pull_request_id = data['github_pull_request_id']
        self.github_url = data['github_url']
        self.id = data['id']
        self.major = data['major']
        self.repository = data['repository']
        self.title = data['title']
        self.type = data['type']
        self.url = data['url']
        self.github_user = GithubUser(data['github_user']) if 'github_user' in data else None
        self.message = data.get('message', '')
        self.message_html = data.get('message_html', '')

    def __repr__(self):
        return prettify(self, 'title', 'major', 'created_at')


class GithubUser:
    """
    **Attributes**

    display_name: :class:`str`

    github_url: :class:`str` or :class:`NoneType`

    id: :class:`int` or :class:`NoneType`

    osu_username: :class:`str` or :class:`NoneType`

    user_id: :class:`int` or :class:`NoneType`

    user_url: :class:`str` or :class:`NoneType`
    """
    __slots__ = (
        "display_name", "github_url", "id", "osu_username",
        "user_id", "user_url"
    )

    def __init__(self, data):
        self.display_name = data['display_name']
        self.github_url = data['github_url']
        self.id = data['id']
        self.osu_username = data['osu_username']
        self.user_id = data['user_id']
        self.user_url = data['user_url']

    def __repr__(self):
        return prettify(self, 'display_name')



"""
Note: Docstrings in this file may be outdated due to laziness of not copying
over docstring edits in the osu/client.py file.
"""
from .http import AsynchronousHTTPHandler as HTTPHandler
from ..objects import *
from ..path import Path
from ..enums import *
from ..auth import AuthHandler
from ..util import parse_mods_arg, parse_enum_args, BeatmapsetSearchFilter
from typing import Union, Optional, Sequence, Dict
from datetime import datetime
from osrparse import Replay


class AsynchronousClient:
    """
    Main object for interacting with osu!api

    **Init Parameters**

    auth: :class:`AuthHandler`
        The AuthHandler object passed in when initiating the Client object

    request_wait_time: Optional[:class:`float`]
        Default is 1.

        This defines the amount of time that the client should wait before making another request.
        It can make it easier to stay within the rate limits without using all your requests up quickly
        and then waiting forever to make another. It's most applicable in bot-type apps.

    limit_per_minute: Optional[:class:`float`]
        Default is 60 because that's the limit peppy requests that we stay under.

        This sets a cap on the number of requests the client is allowed to make within 1 minute of time.

    Make sure if you are changing the ratelimit handling that you are still following peppy's
    TOU for using the API:

    Use the API for good. Don't overdo it. If in doubt, ask before (ab)using :).
    this section may expand as necessary.

    Current rate limit is set at an insanely high 1200 requests per minute,
    with burst capability of up to 200 beyond that.
    If you require more, you probably fall into the above category of abuse.
    If you are doing more than 60 requests a minute,
    you should probably give peppy a yell.
    """

    def __init__(self, auth=None, request_wait_time: Optional[float] = 1.0,
                 limit_per_minute: Optional[float] = 60.0):
        self.auth = auth
        self.http = HTTPHandler(self, request_wait_time, limit_per_minute)

    @classmethod
    def from_client_credentials(cls, client_id: int, client_secret: str, redirect_url: str,
                                scope: Optional[Scope] = Scope.default(), code: Optional[str] = None,
                                request_wait_time: Optional[float] = 1.0,
                                limit_per_minute: Optional[float] = 60.0):
        """
        Returns a :class:`Client` object from client id, client secret, redirect uri, and scope.

        **Parameters**

        client_id: :class:`int`
            API Client id

        client_secret: :class:`int`
            API Client secret

        redirect_uri: :class:`str`
            API redirect uri

        scope: Optional[:class:`Scope`]
            Scopes to use. Default is Scope.default() which is just the public scope.

        code: Optional[:class:`str`]
            If provided, is used to authorize. Read more about this under :class:`AuthHandler.get_auth_token`

        request_wait_time: Optional[:class:`float`]
            Read under Client init parameters.

        limit_per_minute: Optional[:class:`float`]
            Read under Client init parameters.

        **Returns**

        :class:`Client`
        """
        auth = AuthHandler(client_id, client_secret, redirect_url, scope)
        auth.get_auth_token(code)
        return cls(auth, request_wait_time, limit_per_minute)

    async def lookup_beatmap(self, checksum: Optional[str] = None, filename: Optional[str] = None,
                             id: Optional[int] = None) -> Beatmap:
        """
        Returns beatmap.

        Requires OAuth and scope public

        **Parameters**

        checksum: Optional[:class:`str`]
            An md5 beatmap checksum.

        filename: Optional[:class:`str`]
            A filename to lookup

        id: Optional[:class:`int`]
            A beatmap ID to lookup

        **Returns**

        :class:`Beatmap`
        """
        return Beatmap(
            await self.http.make_request('get', Path.beatmap_lookup(), checksum=checksum, filename=filename, id=id))

    async def get_user_beatmap_score(self, beatmap: int, user: int, mode: Optional[Union[str, GameModeStr]] = None,
                                     mods: Optional[Sequence[str]] = None) -> BeatmapUserScore:
        """
        Returns a user's score on a Beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        user: :class:`int`
            Id of the user

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Sequence[:class:`str`]]
            An array of matching mods, or none. Currently doesn't do anything.

        **Returns**

        :class:`BeatmapUserScore`
        """
        mode = parse_enum_args(mode)
        return BeatmapUserScore(
            await self.http.make_request('get', Path.user_beatmap_score(beatmap, user), mode=mode, mods=mods))

    async def get_user_beatmap_scores(self, beatmap: int, user: int,
                                      mode: Optional[Union[str, GameModeStr]] = None) -> Sequence[LegacyScore]:
        """
        Returns a user's scores on a Beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        user: :class:`int`
            Id of the user

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        **Returns**

        Sequence[:class:`LegacyScore`]
        """
        mode = parse_enum_args(mode)
        return list(map(LegacyScore, (await self.http.make_request('get', Path.user_beatmap_scores(beatmap, user),
                                                             mode=mode))["scores"]))

    async def get_beatmap_scores(self, beatmap: int, mode: Optional[Union[str, GameModeStr]] = None,
                                 mods: Optional[Sequence[str]] = None,
                                 type: Optional[Sequence[str]] = None) -> BeatmapScores:
        """
        Returns the top scores for a beatmap

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Sequence[:class:`str`]]
            An array of matching mods, or none. Currently doesn't do anything.

        type: Optional[Sequence[:class:`str`]]
            Beatmap score ranking type. Currently doesn't do anything.

        **Returns**

        :class:`BeatmapScores`
            :class:`LegacyScore` object inside includes "user" and the included user includes "country" and "cover".
        """
        mode = parse_enum_args(mode)
        return BeatmapScores(
            await self.http.make_request('get', Path.beatmap_scores(beatmap), mode=mode, mods=mods, type=type))

    async def get_lazer_beatmap_scores(self, beatmap: int, mode: Optional[Union[str, GameModeStr]] = None,
                                       mods: Optional[Sequence[str]] = None,
                                       type: Optional[Sequence[str]] = None) -> BeatmapScores:
        """
        Returns the top scores for a beatmap on the lazer client.

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Id of the beatmap

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            The game mode to get scores for

        mods: Optional[Sequence[:class:`str`]]
            An array of matching mods, or none. Currently doesn't do anything.

        type: Optional[Sequence[:class:`str`]]
            Beatmap score ranking type. Currently doesn't do anything.

        **Returns**

        :class:`BeatmapScores`
        """
        mode = parse_enum_args(mode)
        return BeatmapScores(await self.http.make_request('get', Path.lazer_beatmap_scores(beatmap), mode=mode,
                                                          mods=mods, type=type), "lazer")

    async def get_beatmap(self, beatmap: int) -> Beatmap:
        """
        Gets beatmap data for the specified beatmap ID.

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            The ID of the beatmap

        **Returns**

        :class:`Beatmap`
            Includes attributes beatmapset, failtimes, and max_combo
        """
        return Beatmap(await self.http.make_request('get', Path.beatmap(beatmap)))

    async def get_beatmaps(self, ids: Optional[Sequence[int]] = None) -> Sequence[Beatmap]:
        """
        Returns list of beatmaps.

        Requires OAuth and scope public

        **Parameters**

        ids: Optional[List[:class:`int`]]
            Beatmap id to be returned. Specify once for each beatmap id requested.
            Up to 50 beatmaps can be requested at once.

        **Returns**

        Sequence[:class:`BeatmapCompact`]
            Includes: beatmapset (with ratings), failtimes, max_combo.
        """
        results = await self.http.make_request('get', Path.beatmaps(), **{"ids[]": ids})
        return list(map(Beatmap, results['beatmaps'])) if results else []

    async def get_beatmap_attributes(self, beatmap: int,
                                     mods: Optional[Union[int, Mods, Sequence[Union[str, Mods, int]]]] = None,
                                     ruleset: Optional[Union[str, GameModeStr]] = None,
                                     ruleset_id: Optional[
                                         Union[int, GameModeInt]] = None) -> BeatmapDifficultyAttributes:
        """
        Returns difficulty attributes of beatmap with specific mode and mods combination.

        Requires OAuth and scope public

        **Parameters**

        beatmap: :class:`int`
            Beatmap id.

        mods: Optional[Union[:class:`int`, Sequence[Union[:class:`str`, :class:`Mods`, :class:`int`]], :class:`Mods`]]
            Mod combination. Can be either a bitset of mods, a Mods enum, or array of any. Defaults to no mods.
            Some mods may cause the api to throw an HTTP 422 error depending on the map's gamemode.

        ruleset: Optional[Union[:class:`GameModeStr`, :class:`int`]]
            Ruleset of the difficulty attributes.
            Only valid if it's the beatmap ruleset or the beatmap can be converted to the specified ruleset.
            Defaults to ruleset of the specified beatmap.

        ruleset_id: Optional[Union[:class:`GameModeInt`, :class:`int`]]
            The same as ruleset but in integer form.

        **Returns**

        :class:`BeatmapDifficultyAttributes`
        """
        ruleset, ruleset_id = parse_enum_args(ruleset, ruleset_id)
        return BeatmapDifficultyAttributes(
            await self.http.make_request('post', Path.get_beatmap_attributes(beatmap), mods=parse_mods_arg(mods),
                                         ruleset=ruleset, ruleset_id=ruleset_id))

    async def get_beatmapset_discussion_posts(self, beatmapset_discussion_id: Optional[int] = None,
                                              limit: Optional[int] = None,
                                              page: Optional[int] = None, sort: Optional[str] = None,
                                              user: Optional[int] = None,
                                              with_deleted: Optional[str] = None) -> dict:
        """
        Returns the posts of the beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmapset_discussion_id: Optional[:class:`int`]
            id of the BeatmapsetDiscussion

        limit: Optional[:class:`int`]
            Maximum number of results

        page: Optional[:class:`int`]
            Search results page.

        sort: Optional[:class:`str`]
            id_desc for newest first; id_asc for oldest first. Defaults to id_desc

        user: Optional[:class:`int`]
            The id of the User

        with_deleted: Optional[:class:`str`]
            The param has no effect as api calls do not currently receive group permissions

        **Returns**

        :class:`dict`
            {
            beatmapsets: :class:`BeatmapsetCompact`,

            cursor: :class:`dict`,

            posts: Sequence[:class:`BeatmapsetDiscussionPost`],

            users: :class:`UserCompact`
            }
        """
        # TODO: Change is supposed to occur on the response given back from the server,
        #  make sure to change it when that happens.
        resp = await self.http.make_request('get', Path.beatmapset_discussion_posts(),
                                            beatmapset_discussion_id=beatmapset_discussion_id,
                                            limit=limit, page=page, sort=sort, user=user, with_deleted=with_deleted)
        return {
            'beatmapsets': list(map(BeatmapsetCompact, resp['beatmapsets'])),
            'cursor': resp['cursor'],
            'posts': list(map(BeatmapsetDiscussionPost, resp['posts'])),
            'users': list(map(UserCompact, resp['users']))
        }

    async def get_beatmapset_discussion_votes(self, beatmapset_discussion_id: Optional[int] = None,
                                              limit: Optional[int] = None,
                                              page: Optional[int] = None, receiver: Optional[int] = None,
                                              score: Optional[int] = None,
                                              sort: Optional[str] = None, user: Optional[int] = None,
                                              with_deleted: Optional[str] = None) -> dict:
        """
        Returns the votes given to beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmapset_discussion_id: Optional[:class:`int`]
            id of the BeatmapsetDiscussion

        limit: Optional[:class:`int`]
            Maximum number of results

        page: Optional[:class:`int`]
            Search results page.

        receiver: Optional[:class:`int`]
            The id of the User receiving the votes.

        score: Optional[:class:`int`]
            1 for upvote, -1 for downvote

        sort: Optional[:class:`str`]
            id_desc for newest first; id_asc for oldest first. Defaults to id_desc

        user: Optional[:class:`int`]
            The id of the User giving the votes.

        with_deleted: Optional[:class:`str`]
            The param has no effect as api calls do not currently receive group permissions

        **Returns**

        :class:`dict`
            {
            cursor: :class:`dict`,

            discussions: Sequence[:class:`BeatmapsetDiscussion`],

            users: Sequence[:class:`UserCompact`],

            votes: Sequence[:class:`BeatmapsetDiscussionVote`]
            }
        """
        # TODO: Change is supposed to occur on the response given back from the server,
        #  make sure to change it when that happens.
        resp = await self.http.make_request('get', Path.beatmapset_discussion_votes(),
                                            beatmapset_discussion_id=beatmapset_discussion_id,
                                            limit=limit, receiver=receiver, score=score, page=page, sort=sort,
                                            user=user,
                                            with_deleted=with_deleted)
        return {
            'cursor': resp['cursor'],
            'discussions': list(map(BeatmapsetDiscussion, resp['discussions'])),
            'users': list(map(UserCompact, resp['users'])),
            'votes': list(map(BeatmapsetDiscussionVote, resp['votes']))
        }

    async def get_beatmapset(self, beatmapset_id: int) -> Beatmapset:
        """
        Get beatmapset by id.

        Requires OAuth and scope public

        **Parameters**

        beatmapset_id: :class:`int`

        **Returns**

        :class:`Beatmapset`
        """
        return Beatmapset(await self.http.make_request('get', Path.get_beatmapset(beatmapset_id)))

    async def get_beatmapset_discussions(self, beatmap_id: Optional[int] = None, beatmapset_id: Optional[int] = None,
                                         beatmapset_status: Optional[str] = None, limit: Optional[int] = None,
                                         message_types: Optional[Sequence[str]] = None,
                                         only_unresolved: Optional[bool] = None,
                                         page: Optional[int] = None, sort: Optional[str] = None,
                                         user: Optional[int] = None,
                                         with_deleted: Optional[str] = None) -> dict:
        """
        Returns a list of beatmapset discussions

        Requires OAuth and scope public

        **Parameters**

        beatmap_id: Optional[:class:`int`]
            id of the Beatmap

        beatmapset_id: Optional[:class:`int`]
            id of the Beatmapset

        beatmapset_status: Optional[:class:`str`]
            One of all, ranked, qualified, disqualified, never_qualified. Defaults to all.

        limit: Optional[:class:`int`]
            Maximum number of results.

        message_types: Optional[Sequence[:class:`str`]]
            suggestion, problem, mapper_note, praise, hype, review. Blank defaults to all types.

        only_unresolved: Optional[:class:`bool`]
            true to show only unresolved issues; false, otherwise. Defaults to false.

        page: Optional[:class:`int`]
            Search result page.

        sort: Optional[:class:`str`]
            id_desc for newest first; id_asc for oldest first. Defaults to id_desc.

        user: Optional[:class:`int`]
            The id of the User.

        with_deleted: Optional[:class:`str`]
            This param has no effect as api calls do not currently receive group permissions.

        **Returns**

        :class:`dict`
            {

            beatmaps: Sequence[:class:`Beatmap`],
                List of beatmaps associated with the discussions returned.

            cursor: :class:`dict`,

            discussions: Sequence[:class:`BeatmapsetDiscussion`],
                List of discussions according to sort order.

            included_discussions: Sequence[:class:`BeatmapsetDiscussion`],
                Additional discussions related to discussions.

            reviews_config.max_blocks: :class:`int`,
                Maximum number of blocks allowed in a review.

            users: Sequence[:class:`UserCompact`]
                List of users associated with the discussions returned.

            }
        """
        # TODO: Change is supposed to occur on the response given back from the server,
        #  make sure to change it when that happens.
        message_types = {"message_types[]": message_types}
        resp = await self.http.make_request('get', Path.beatmapset_discussions(), beatmap_id=beatmap_id,
                                            beatmapset_id=beatmapset_id,
                                            beatmapset_status=beatmapset_status, limit=limit,
                                            only_unresolved=only_unresolved,
                                            page=page, sort=sort, user=user, with_deleted=with_deleted,
                                            **message_types)
        return {
            'beatmaps': list(map(Beatmap, resp['beatmaps'])),
            'cursor': resp['cursor'],
            'discussions': list(map(BeatmapsetDiscussion, resp['discussions'])),
            'included_discussions': list(map(BeatmapsetDiscussion, resp['included_discussions'])),
            'reviews_config.max_blocks': resp['reviews_config'],
            'users': list(map(UserCompact, resp['users']))
        }

    async def get_changelog_build(self, stream: str, build: str) -> Build:
        """
        Returns details of the specified build.

        **Parameters**

        stream: :class:`str`
            Update stream name.

        build: :class:`str`
            Build version.

        **Returns**

        A :class:`Build` with changelog_entries, changelog_entries.github_user, and versions included.
        """
        return Build(await self.http.make_request('get', Path.get_changelog_build(stream, build)))

    async def get_changelog_listing(self, from_version: Optional[str] = None, max_id: Optional[int] = None,
                                    stream: Optional[str] = None, to: Optional[str] = None,
                                    message_formats: Optional[Sequence[str]] = None) -> \
            Dict[str, Union[Sequence[Build], Sequence[UpdateStream], Dict[str, Union[str, int, None]]]]:
        """
        Returns a listing of update streams, builds, and changelog entries.

        **Parameters**

        from_version: Optional[:class:`str`]
            Minimum build version.

        max_id: Optional[:class:`int`]
            Maximum build ID.

        stream: Optional[:class:`str`]
            Stream name to return builds from.

        to: Optional[:class:`str`]
            Maximum build version.

        message_formats: Optional[Sequence[:class:`str`]]
            html, markdown. Default to both.

        **Returns**

        {

        "build": Sequence[:class:`Build`]

        "search": {

            "from": :class:`str`
                from_version input.

            "limit": :class:`int`
                Always 21.

            "max_id": :class:`int`
                max_id input.

            "stream": :class:`str`
                stream input.

            "to": :class:`str`
                to input.

        }

        "streams": Sequence[:class:`UpdateStream`]

        }
        """
        response = await self.http.make_request('get', Path.get_changelog_listing(), max_id=max_id, stream=stream,
                                                to=to, message_formats=message_formats, **{"from": from_version})
        return {
            "build": list(map(Build, response['builds'])),
            "search": response['search'],
            "streams": list(map(UpdateStream, response['streams'])),
        }

    async def lookup_changelog_build(self, changelog: str, key: Optional[str] = None,
                                     message_formats: Optional[Sequence[str]] = None) -> Build:
        """
        Returns details of the specified build.

        **Parameter**

        changelog: :class:`str`
            Build version, update stream name, or build ID.

        key: Optional[:class:`str`]
            Unset to query by build version or stream name, or id to query by build ID.

        message_formats: Optional[Sequence[:class:`str`]]
            html, markdown. Default to both.

        **Returns**

        A :class:`Build` with changelog_entries, changelog_entries.github_user, and versions included.
        """
        return Build(await self.http.make_request('get', Path.lookup_changelog_build(changelog), key=key,
                                                  message_formats=message_formats))

    async def create_new_pm(self, target_id: int, message: str, is_action: bool) -> dict:
        """
        This endpoint allows you to create a new PM channel.

        Requires OAuth and scope chat.write

        **Parameters**

        target_id: :class:`int`
            user_id of user to start PM with

        message: :class:`str`
            message to send

        is_action: :class:`bool`
            whether the message is an action

        **Returns**

        :class:`dict`
            {

            new_channel_id: :class:`int`
                channel_id of newly created ChatChannel

            presence: Sequence[:class:`ChatChannel`]
                array of ChatChannel

            message: :class:`ChatMessage`
                the sent ChatMessage

            }
        """
        data = {'target_id': target_id, 'message': message, 'is_action': is_action}
        resp = await self.http.make_request('post', Path.create_new_pm(), data=data)
        return {
            'new_channel_id': resp['new_channel_id'],
            'presence': list(map(ChatChannel, resp['presence'])),
            'message': ChatMessage(resp['message'])
        }

    async def get_updates(self, since: int, channel_id: Optional[int] = None, limit: Optional[int] = None) -> dict:
        """
        This endpoint returns new messages since the given message_id along with updated channel 'presence' data.

        Requires OAuth and scope lazer

        **Parameters**

        since: :class:`int`
            The message_id of the last message to retrieve messages since

        channel_id: Optional[:class:`int`]
            If provided, will only return messages for the given channel

        limit: Optional[:class:`int`]
            number of messages to return (max of 50)

        **Returns**

        :class:`dict`
            {
            presence: List[:class:`ChatChannel`],

            messages: List[:class:`ChatMessage`]

            }
        """
        resp = await self.http.make_request('post', Path.get_updates(), since=since,
                                            channel_id=channel_id, limit=limit)
        return {
            'presence': list(map(ChatChannel, resp['presence'])),
            'messages': list(map(ChatMessage, resp['messages'])),
            'silences': list(map(UserSilence, resp['silences']))
        }

    async def get_channel_messages(self, channel_id: int, limit: Optional[int] = None, since: Optional[int] = None,
                                   until: Optional[int] = None) -> Sequence[ChatMessage]:
        """
        This endpoint returns the chat messages for a specific channel.

        Requires OAuth and scope lazer

        **Parameter**

        channel_id: :class:`int`
            The ID of the channel to retrieve messages for

        limit: Optional[:class:`int`]
            number of messages to return (max of 50)

        since: Optional[:class:`int`]
            messages after the specified message id will be returned

        until: Optional[:class:`int`]
            messages up to but not including the specified message id will be returned

        **Returns**

        Sequence[:class:`ChatMessage`]
            list containing :class:`ChatMessage` objects
        """
        return list(map(ChatMessage,
                        await self.http.make_request('post', Path.get_channel_messages(channel_id), limit=limit,
                                                     since=since,
                                                     until=until)))

    async def send_message_to_channel(self, channel_id: int, message: str, is_action: bool) -> ChatMessage:
        """
        This endpoint sends a message to the specified channel.

        Requires OAuth and scope lazer

        **Parameters**

        channel_id: :class:`int`
            The channel_id of the channel to send message to

        message: :class:`str`
            message to send

        is_action: :class:`bool`
            whether the message is an action

        **Returns**

        :class:`ChatMessage`
        """
        data = {'message': message, 'is_action': is_action}
        return ChatMessage(await self.http.make_request('post', Path.send_message_to_channel(channel_id), data=data))

    async def join_channel(self, channel: int, user: int) -> ChatChannel:
        """
        This endpoint allows you to join a public channel.

        Requires OAuth and scope lazer

        **Parameters**

        channel: :class:`int`

        user: :class:`int`

        **Returns**

        :class:`ChatChannel`
        """
        return ChatChannel(await self.http.make_request('put', Path.join_channel(channel, user)))

    async def leave_channel(self, channel: int, user: int):
        """
        This endpoint allows you to leave a public channel.

        Requires OAuth and scope lazer

        **Parameters**

        channel: :class:`int`

        user: :class:`int`
        """
        await self.http.make_request('delete', Path.leave_channel(channel, user))

    async def mark_channel_as_read(self, channel: str, message: str, channel_id: int, message_id: int):
        """
        This endpoint marks the channel as having being read up to the given message_id.

        Requires OAuth and scope lazer

        **Parameters**

        channel: :class:`str`

        message: :class:`str`

        channel_id: :class:`int`
            The channel_id of the channel to mark as read

        message_id: :class:`int`
            The message_id of the message to mark as read up to
        """
        await self.http.make_request('put', Path.mark_channel_as_read(channel, message), channel_id=channel_id,
                                     message_id=message_id)

    async def get_channel_list(self) -> Sequence[ChatChannel]:
        """
        This endpoint returns a list of all joinable public channels.

        Requires OAuth and scope lazer

        **Returns**

        Sequence[:class:`ChatChannel`]
        """
        return list(map(ChatChannel, await self.http.make_request('get', Path.get_channel_list())))

    async def create_channel(self, type: str, target_id: Optional[int] = None) -> ChatChannel:
        """
        This endpoint creates a new channel if doesn't exist and joins it.
        Currently only for rejoining existing PM channels which the user has left.

        Requires OAuth and scope lazer

        **Parameter**

        type: :class:`str`
            channel type (currently only supports "PM")

        target_id: Optional[:class:`int`]
            target user id for type PM

        **Returns**

        :class:`ChatChannel`
             contains recent_messages attribute. Note that if there's no existing PM channel,
             most of the fields will be blank. In that case, send a message (create_new_pm)
             instead to create the channel.
        """
        data = {'type': type, 'target_id': target_id}
        return ChatChannel(await self.http.make_request('post', Path.create_channel(), data=data))

    async def get_channel(self, channel: int) -> dict:
        """
        Gets details of a chat channel.

        Requires OAuth and scope lazer

        **Parameter**

        channel: :class:`int`

        **Returns**

        :class:`dict`
            {
            channel: :class:`ChatChannel`,

            users: :class:`UserCompact`

            }
        """
        resp = await self.http.make_request('get', Path.get_channel(channel))
        return {
            'channel': ChatChannel(resp['channel']),
            'users': UserCompact(resp['users']),
        }

    async def get_comments(self, commentable_type: Optional[str] = None, commentable_id: Optional[int] = None,
                           cursor: Optional[dict] = None, parent_id: Optional[int] = None,
                           sort: Optional[Union[str, CommentSort]] = None) -> CommentBundle:
        """
        Returns a list comments and their replies up to 2 levels deep.

        Does not require OAuth

        **Parameter**

        commentable_type: Optional[:class:`str`]
            The type of resource to get comments for.

        commentable_id: Optional[:class:`int`]
            The id of the resource to get comments for.

        cursor: Optional[:class:`dict`]
            Pagination option. See :ref:`CommentSort` for detail.
            The format follows Cursor except it's not currently included in the response.

        parent_id: Optional[:class:`int`]
            Limit to comments which are reply to the specified id. Specify 0 to get top level comments.

        sort: Optional[Union[:class:`str`, :class:`CommentSort`]]
            Sort option as defined in :ref:`CommentSort`.
            Defaults to new for guests and user-specified default when authenticated.

        **Returns**

        :class:`CommentBundle`
            pinned_comments is only included when commentable_type and commentable_id are specified.
        """
        sort = parse_enum_args(sort)
        return CommentBundle(await self.http.make_request('get', Path.get_comments(),
                                                          commentable_type=commentable_type,
                                                          commentable_id=commentable_id, **(cursor if cursor else {}),
                                                          parent_id=parent_id, sort=sort))

    async def post_comment(self, commentable_id: Optional[int] = None, commentable_type: Optional[str] = None,
                           message: Optional[str] = None, parent_id: Optional[int] = None) -> CommentBundle:
        """
        Posts a new comment to a comment thread.

        Requires OAuth and scope lazer

        **Parameter**

        commentable_id: Optional[:class:`int`]
            Resource ID the comment thread is attached to

        commentable_type: Optional[:class:`str`]
            Resource type the comment thread is attached to

        message: Optional[:class:`str`]
            Text of the comment

        parent_id: Optional[:class:`int`]
            The id of the comment to reply to, null if not a reply

        **Returns**

        :class:`CommentBundle`
        """
        params = {
            'comment.commentable_id': commentable_id,
            'comment_commentable_type': commentable_type,
            'comment.message': message,
            'comment.parent_id': parent_id
        }
        return CommentBundle(await self.http.make_request('post', Path.post_new_comment(), params=params))

    async def get_comment(self, comment: int) -> CommentBundle:
        """
        Gets a comment and its replies up to 2 levels deep.

        Does not require OAuth

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(await self.http.make_request('get', Path.get_comment(comment)))

    async def edit_comment(self, comment: int, message: Optional[str] = None) -> CommentBundle:
        """
        Edit an existing comment.

        Requires OAuth and scope lazer

        **Parameters**

        comment: :class:`int`
            Comment id

        message: Optional[:class:`str`]
            New text of the comment

        **Returns**

        :class:`CommentBundle`
        """
        params = {'comment.message': message}
        return CommentBundle(await self.http.make_request('patch', Path.edit_comment(comment), params=params))

    async def delete_comment(self, comment: int) -> CommentBundle:
        """
        Deletes the specified comment.

        Requires OAuth and scope lazer

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(await self.http.make_request('delete', Path.delete_comment(comment)))

    async def add_comment_vote(self, comment: int) -> CommentBundle:
        """
        Upvotes a comment.

        Requires OAuth and scope lazer

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(await self.http.make_request('post', Path.add_comment_vote(comment)))

    async def remove_comment_vote(self, comment: int) -> CommentBundle:
        """
        Un-upvotes a comment.

        Requires OAuth and scope lazer

        **Parameters**

        comment: :class:`int`
            Comment id

        **Returns**

        :class:`CommentBundle`
        """
        return CommentBundle(await self.http.make_request('delete', Path.remove_comment_vote(comment)))

    async def reply_topic(self, topic: int, body: str) -> ForumPost:
        """
        Create a post replying to the specified topic.

        Requires OAuth and scope forum.write

        **Parameters**

        topic: :class:`int`
            Id of the topic to be replied to.

        body: :class:`str`
            Content of the reply post.

        **Returns**

        :class:`ForumPost`
            body attributes included
        """
        data = {'body': body}
        return ForumPost(await self.http.make_request('post', Path.reply_topic(topic), data=data))

    async def create_topic(self, body: str, forum_id: int, title: str, with_poll: Optional[bool] = None,
                           hide_results: Optional[bool] = None, length_days: Optional[int] = None,
                           max_options: Optional[int] = None, poll_options: Optional[str] = None,
                           poll_title: Optional[str] = None, vote_change: Optional[bool] = None) -> dict:
        """
        Create a new topic.

        Requires OAuth and scope forum.write

        **Parameters**

        body: :class:`str`
            Content of the topic.

        forum_id: :class:`int`
            Forum to create the topic in.

        title: :class:`str`
            Title of the topic.

        with_poll: Optional[:class:`bool`]
            Enable this to also create poll in the topic (default: false).

        hide_results: Optional[:class:`bool`]
            Enable this to hide result until voting period ends (default: false).

        length_days: Optional[:class:`int`]
            Number of days for voting period. 0 means the voting will never ends (default: 0).
            This parameter is required if hide_results option is enabled.

        max_options: Optional[:class:`int`]
            Maximum number of votes each user can cast (default: 1).

        poll_options: Optional[:class:`str`]
            Newline-separated list of voting options. BBCode is supported.

        poll_title: Optional[:class:`str`]
            Title of the poll.

        vote_change: Optional[:class:`bool`]
            Enable this to allow user to change their votes (default: false).

        **Returns**

        :class:`dict`
            {
            topic: :class:`ForumTopic`

            post: :class:`ForumPost`
                includes body

            }
        """
        data = {'body': body, 'forum_id': forum_id, 'title': title, 'with_poll': with_poll}
        if with_poll:
            if poll_options is None or poll_title is None:
                raise TypeError("poll_options and poll_title are required since the topic has a poll.")
            data.update({'forum_topic_poll': {
                'hide_results': hide_results, 'length_days': length_days,
                'max_options': max_options, 'poll_options': poll_options,
                'poll_title': poll_title, 'vote_change': vote_change
            }})
        resp = await self.http.make_request('post', Path.create_topic(), data=data)
        return {
            'topic': ForumTopic(resp['topic']),
            'post': ForumPost(resp['post'])
        }

    async def get_topic_and_posts(self, topic: int, cursor: Optional[dict] = None, sort: Optional[str] = None,
                                  limit: Optional[int] = None, start: Optional[int] = None,
                                  end: Optional[int] = None) -> dict:
        """
        Get topic and its posts.

        Requires OAuth and scope public

        **Parameters**

        topic: :class:`int`
            Id of the topic.

        cursor: Optional[:class:`dict`]
            To be used to fetch the next page of results

        sort: Optional[:class:`str`]
            Post sorting option. Valid values are id_asc (default) and id_desc.

        limit: Optional[:class:`int`]
            Maximum number of posts to be returned (20 default, 50 at most).

        start: Optional[:class:`int`]
            First post id to be returned with sort set to id_asc. This parameter is ignored if cursor is specified.

        end: Optional[:class:`int`]
            First post id to be returned with sort set to id_desc. This parameter is ignored if cursor is specified.

        **Returns**

        :class:`dict`
            {
            cursor: :class:`dict`,

            search: :class:`dict`,

            posts: Sequence[:class:`ForumPost`],

            topic: :class:`ForumTopic`

            }
        """
        resp = await self.http.make_request('get', Path.get_topic_and_posts(topic), **(cursor if cursor else {}),
                                            sort=sort,
                                            limit=limit, start=start, end=end)
        return {
            'cursor': resp['cursor'],
            'search': resp['search'],
            'posts': list(map(ForumPost, resp['posts'])),
            'topic': ForumTopic(resp['topic'])
        }

    async def edit_topic(self, topic: int, topic_title: str) -> ForumTopic:
        """
        Edit topic. Only title can be edited through this endpoint.

        Requires OAuth and scope forum.write

        **Parameters**

        topic: :class:`int`
            Id of the topic.

        topic_title: :class:`str`
            New topic title.

        **Returns**

        :class:`ForumTopic`
        """
        data = {'forum_topic': {'topic_title': topic_title}}
        return ForumTopic(await self.http.make_request('patch', Path.edit_topic(topic), data=data))

    async def edit_post(self, post: int, body: str) -> ForumPost:
        """
        Edit specified forum post.

        Requires OAuth and scope forum.write

        post: :class:`int`
            Id of the post.

        body: :class:`str`
            New post content in BBCode format.

        **Returns**

        :class:`ForumPost`
        """
        data = {'body': body}
        return ForumPost(await self.http.make_request('patch', Path.edit_post(post), data=data))

    async def search(self, mode: Optional[Union[str, WikiSearchMode]] = None, query: Optional[str] = None,
                     page: Optional[int] = None) -> dict:
        """
        Searches users and wiki pages.

        Requires OAuth and scope public

        **Parameters**

        mode: Optional[Union[:class:`str`, :class:`WikiSearchMode`]]
            Either all, user, or wiki_page. Default is all.

        query: Optional[:class:`str`]
            Search keyword.

        page: Optional[:class:`int`]
            Search result page. Ignored for mode all.

        **Returns**

        :class:`dict`
            {

            user: :class:`dict`
                For all or user mode. Only first 100 results are accessible
                {
                results: Sequence

                total: :class:`int`
                }

            wiki_page: :class:`dict`
                For all or wiki_page mode
                {
                results: Sequence

                total: :class:`int`
                }

            }
        """
        mode = parse_enum_args(mode)
        resp = await self.http.make_request('get', Path.search(), mode=mode, query=query, page=page)
        return {
            'user': {'results': resp['user']['data'],
                     'total': resp['user']['total']} if mode is None or mode == 'all' or mode == 'user' else None,
            'wiki_page': {'results': resp['wiki_page']['data'], 'total': resp['wiki_page'][
                'total']} if mode is None or mode == 'all' or mode == 'wiki_page' else None
        }

    async def get_user_highscore(self, room: int, playlist: int, user: int) -> MultiplayerScore:
        """
        Requires OAuth and scope lazer

        **Parameters**

        room: :class:`int`
            Id of the room.

        playlist: :class:`int`
            Id of the playlist item.

        user: :class:`int`
            User id.

        **Returns**

        :class:`MultiplayerScores`
        """
        return MultiplayerScore(await self.http.make_request('get', Path.get_user_high_score(room, playlist, user)))

    async def get_scores(self, room: int, playlist: int, limit: Optional[int] = None,
                         sort: Optional[Union[str, MultiplayerScoresSort]] = None,
                         cursor: Optional[dict] = None) -> MultiplayerScores:
        """
        Requires OAuth and scope public

        **Parameters**

        room: :class:`int`
            Id of the room.

        playlist: :class:`int`
            Id of the playlist item.

        limit: Optional[:class:`int`]
            Number of scores to be returned.

        sort: Optional[Union[:class:`str`, :class:`MultiplayerScoresSort`]]
            :ref:`MultiplayerScoresSort` parameter.

        cursor: Optional[:class:`dict`]

        **Returns**

        :class:`MultiplayerScores`
        """
        sort = parse_enum_args(sort)
        return MultiplayerScores(
            await self.http.make_request('get', Path.get_scores(room, playlist), limit=limit, sort=sort,
                                         **(cursor if cursor else {})))

    async def get_score(self, room: int, playlist: int, score: int) -> MultiplayerScore:
        """
        Requires OAuth and scope lazer

        **Parameters**

        room: :class:`int`
            Id of the room.

        playlist: :class:`int`
            Id of the playlist item.

        score: :class:`int`
            Id of the score.

        **Returns**

        :class:`MultiplayerScore`
        """
        return MultiplayerScore(await self.http.make_request('get', Path.get_score(room, playlist, score)))

    async def get_news_listing(self, limit: Optional[int] = None, year: Optional[int] = None,
                               cursor: Optional[dict] = None) -> dict:
        """
        Returns a list of news posts and related metadata.

        **Parameters**

        limit: Optional[:class:`int`]
            Maximum number of posts (12 default, 1 minimum, 21 maximum).

        year: Optional[:class:`int`]
            Year to return posts from.

        cursor: Optional[:class:`dict`]
            Cursor for pagination.

        **Returns**

        {

        cursor: :class:`dict`

        news_posts: Sequence[:class:`NewsPost`]
            Includes preview.

        news_sidebar: {

            current_year: :class:`int`
                Year of the first post's publish time, or current year if no posts returned.

            years: :class:`int`
                All years during which posts have been published.

            news_posts: Sequence[:class:`NewsPost`]
                All posts published during current_year.

        }

        search: {

            limit: :class:`int`
                Clamped limit input.

            sort: :class:`str`
                Always published_desc.

            }

        }
        """
        response = await self.http.make_request('get', Path.get_news_listing(), limit=limit, year=year, cursor=cursor)
        return {
            "cursor": response['cursor'],
            "news_posts": list(map(NewsPost, response["news_posts"])),
            "news_sidebar": {
                "current_year": response['news_sidebar']['current_year'],
                "years": response['news_sidebar']['years'],
                "news_posts": list(map(NewsPost, response['news_sidebar']['news_posts'])),
            },
            "search": response['search']
        }

    async def get_news_post(self, news: str, key: Optional[str] = None) -> NewsPost:
        """
        Returns details of the specified news post.

        **Parameters**

        news: class:`str`
            News post slug or ID.

        key: Optional[:class:`str`]
            Unset to query by slug, or id to query by ID.

        **Returns**

        Returns a :class:`NewsPost` with content and navigation included.
        """
        return NewsPost(await self.http.make_request('get', Path.get_news_post(news), key=key))

    async def get_notifications(self, max_id: Optional[int] = None) -> dict:
        """
        This endpoint returns a list of the user's unread notifications.
        Sorted descending by id with limit of 50.

        Requires OAuth and scope lazer

        **Parameters**

        max_id: Optional[:class:`int`]
            Maximum id fetched. Can be used to load earlier notifications.
            Defaults to no limit (fetch latest notifications)

        **Returns**

        :class:`dict`
            {

            has_more: :class:`bool`,
                whether or not there are more notifications

            notifications: Sequence[:class:`Notification`],

            unread_count: :class:`bool`
                total unread notifications

            notification_endpoint: :class:`str`
                url to connect to websocket server

            }
        """
        resp = await self.http.make_request('get', Path.get_notifications(), max_id=max_id)
        return {
            'has_more': resp['has_more'],
            'notifications': list(map(Notification, resp['notifications'])),
            'unread_count': resp['unread_count'],
            'notification_endpoint': resp['notification_endpoint'],
        }

    async def mark_notifications_read(self, ids: Sequence[int]):
        """
        This endpoint allows you to mark notifications read.

        Requires OAuth and scope lazer

        **Parameters**

        ids: Sequence[:class:`int`]
            ids of notifications to be marked as read.
        """
        data = {'ids[]': ids}
        await self.http.make_request('post', Path.mark_notifications_as_read(), data=data)

    async def revoke_current_token(self):
        """
        Requires OAuth
        """
        await self.http.make_request('delete', self, Path.revoke_current_token())

    async def get_ranking(self, mode: Union[str, GameModeStr], type: Union[str, RankingType],
                          country: Optional[str] = None,
                          cursor: Optional[dict] = None,
                          filter: Optional[str] = None, spotlight: Optional[int] = None,
                          variant: Optional[str] = None) -> Rankings:
        """
        Gets the current ranking for the specified type and game mode.

        Requires OAuth and scope public

        mode: Union[:class:`str`, :class:`GameModeStr`]

        type: Union[:class:`str`, :class:`RankingType`]
            :ref:`RankingType`

        country: Optional[:class:`str`]
            Filter ranking by country code. Only available for type of performance.

        cursor: Optional[:class:`dict`]

        filter: Optional[:class:`str`]
            Either all (default) or friends.

        spotlight: Optional[:class:`int`]
            The id of the spotlight if type is charts.
            Ranking for latest spotlight will be returned if not specified.

        variant: Optional[:class:`str`]
            Filter ranking to specified mode variant. For mode of mania, it's either 4k or 7k.
            Only available for type of performance.

        **Returns**

        :class:`Rankings`
        """
        mode, type = parse_enum_args(mode, type)
        return Rankings(
            await self.http.make_request('get', Path.get_ranking(mode, type), country=country,
                                         **(cursor if cursor else {}),
                                         filter=filter,
                                         spotlight=spotlight, variant=variant))

    async def get_spotlights(self) -> Spotlights:
        """
        Gets the list of spotlights.

        Requires OAuth and scope public

        **Returns**

        :class:`Spotlights`
        """
        return Spotlights(await self.http.make_request('get', Path.get_spotlights()))

    async def get_own_data(self, mode: Union[str, GameModeStr] = "") -> User:
        """
        Similar to get_user but with authenticated user (token owner) as user id.

        Requires OAuth and scope identify

        **Parameters**

        mode: Optional[:class:`str`, :class:`GameModeStr`]
            GameMode. User default mode will be used if not specified.

        **Returns**

        See return for get_user
        """
        mode = parse_enum_args(mode)
        return User(await self.http.make_request('get', Path.get_own_data(mode)))

    async def get_user_kudosu(self, user: int, limit: Optional[int] = None, offset: Optional[int] = None):
        """
        Returns kudosu history.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[:class:`KudosuHistory`]
        """
        return list(
            map(KudosuHistory,
                await self.http.make_request('get', Path.get_user_kudosu(user), limit=limit, offset=offset)))

    async def get_user_scores(self, user: int, type: Union[UserScoreType, str], include_fails: Optional[int] = None,
                        mode: Optional[Union[str, GameModeStr]] = None, limit: Optional[int] = None,
                        offset: Optional[int] = None) -> Sequence[LegacyScore]:
        """
        This endpoint returns the scores of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        type: union[:class:`UserScoreType` :class:`str`]
            Score type. Must be one of these: best, firsts, recent

        include_fails: Optional[:class:`int`]
            Only for recent scores, include scores of failed plays. Set to 1 to include them. Defaults to 0.

        mode: Optional[Union[:class:`GameModeStr`, :class:`str`]]
            game mode of the scores to be returned. Defaults to the specified user's mode.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[:class:`LegacyScore`]
            Includes attributes beatmap, beatmapset, weight: Only for type best, user
        """
        mode, type = parse_enum_args(mode, type)
        return [LegacyScore(score) for score in await self.http.make_request('get', Path.get_user_scores(user, type),
                                                                             include_fails=include_fails, mode=mode,
                                                                             limit=limit, offset=offset)]

    async def get_user_beatmaps(self, user: int, type: Union[str, UserBeatmapType], limit: Optional[int] = None,
                                offset: Optional[int] = None) -> Sequence[Union[BeatmapPlaycount, Beatmapset]]:
        """
        Returns the beatmaps of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        type: Union[:class:`str`, :class:`UserBeatmapType`]
            Beatmap type. Can be one of the following - favourite, graveyard, loved, most_played, pending, ranked.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[Union[:class:`BeatmapPlaycount`, :class:`Beatmapset`]]
            :class:`BeatmapPlaycount` for type most_played or :class:`Beatmapset` for any other type.
        """
        object_type = Beatmapset
        type = parse_enum_args(type)
        if type == 'most_played':
            object_type = BeatmapPlaycount
        return list(map(object_type,
                        await self.http.make_request('get', Path.get_user_beatmaps(user, type), limit=limit,
                                                     offset=offset)))

    async def get_user_recent_activity(self, user: int, limit: Optional[int] = None, offset: Optional[int] = None) -> \
            Sequence[Event]:
        """
        Returns recent activity.

        Requires OAuth and scope public

        **Parameters**

        user: :class:`int`
            Id of the user.

        limit: Optional[:class:`int`]
            Maximum number of results.

        offset: Optional[:class:`int`]
            Result offset for pagination.

        **Returns**

        Sequence[:class:`Event`]
            list of :class:`Event` objects
        """
        return list(
            map(Event,
                await self.http.make_request('get', Path.get_user_recent_activity(user), limit=limit, offset=offset)))

    async def get_user(self, user: int, mode: Optional[Union[str, GameModeStr]] = '',
                       key: Optional[str] = None) -> User:
        """
        This endpoint returns the detail of specified user.

        Requires OAuth and scope public

        **Parameters**

        user: Union[:class:`int`, :class:`str`]
            Id or username of the user. Id lookup is prioritised unless key parameter is specified.
            Previous usernames are also checked in some cases.

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]
            User default mode will be used if not specified.

        key: Optional[:class:`str`]
            Type of user passed in url parameter. Can be either id or username
            to limit lookup by their respective type. Passing empty or invalid
            value will result in id lookup followed by username lookup if not found.

        **Returns**

        :class:`User`
            Includes following attributes: account_history, active_tournament_banner,
            badges, beatmap_playcounts_count, favourite_beatmapset_count, follower_count,
            graveyard_beatmapset_count, groups, loved_beatmapset_count,
            mapping_follower_count, monthly_playcounts, page, pending_beatmapset_count,
            previous_usernames, rank_history, ranked_beatmapset_count, replays_watched_counts,
            scores_best_count, scores_first_count, scores_recent_count, statistics,
            statistics.country_rank, statistics.rank, statistics.variants, support_level,
            user_achievements.
        """
        mode = parse_enum_args(mode)
        return User(await self.http.make_request('get', Path.get_user(user, mode), key=key))

    async def get_users(self, ids: Sequence[int]) -> Sequence[UserCompact]:
        """
        Returns list of users.

        Requires OAuth and scope public

        **Parameters**

        ids: Sequence[:class:`int`]
            User id to be returned. Specify once for each user id requested.
            Up to 50 users can be requested at once.

        **Returns**

        Sequence[:class:`UserCompact`]
            Includes attributes: country, cover, groups, statistics_rulesets.
        """
        res = await self.http.make_request('get', Path.get_users(), **{"ids[]": ids})
        return list(map(UserCompact, res["users"]))

    async def get_wiki_page(self, locale: str, path: str) -> WikiPage:
        """
        The wiki article or image data.

        No OAuth required.

        **Parameters**

        locale: :class:`str`
            Two-letter language code of the wiki page.

        path: :class:`str`
            The path name of the wiki page.

        **Returns**

        :class:`WikiPage`
        """
        return WikiPage(await self.http.make_request('get', Path.get_wiki_page(locale, path)))

    async def get_beatmapset_events(self, page: Optional[int] = None, limit: Optional[int] = None,
                                    sort: Optional[Union[str, BeatmapsetEventSort]] = None,
                                    type: Optional[Union[str, BeatmapsetEventType]] = None,
                                    min_date: Optional[Union[str, datetime]] = None,
                                    max_date: Optional[Union[str, datetime]] = None) -> \
            Dict[str, Union[Sequence[BeatmapsetEvent], Dict, Sequence[UserCompact]]]:
        """
        Returns a list of beatmapset events.

        Requires OAuth and scope public.

        **Parameters**

        page: Optional[:class:`int`]

        limit: Optional[:class:`int`]

        sort: Optional[Union[:class:`str`, :class:`BeatmapsetEventSort`]]
            Specified a sort order.

        type: Optional[Union[:class:`str`, :class:`BeatmapsetEventType`]]
            Specifies for only a certain type of event to be returned.

        **Returns**

        Dict[str, Union[Sequence[BeatmapsetEvent], Dict, Sequence[UserCompact]]]

        {

            'events': Sequence[BeatmapsetEvent],

            'reviews_config': Dict,

            'users': Sequence[UserCompact]

        }
        """
        sort, type = parse_enum_args(sort, type)
        if isinstance(min_date, datetime):
            min_date = min_date.isoformat()
        if isinstance(max_date, datetime):
            max_date = max_date.isoformat()
        resp = await self.http.make_request('get', Path.get_beatmapset_events(), page=page, limit=limit, sort=sort,
                                            type=type, min_date=min_date, max_date=max_date)
        return {
            "events": [BeatmapsetEvent(event) for event in resp['events']],
            "reviews_config": resp['reviewsConfig'],
            "users": [UserCompact(user) for user in resp['users']],
        }

    async def get_matches(self, limit=None, sort=None) -> Dict[str, Union[Sequence[Match], Dict]]:
        """
        Returns a list of matches.

        Requires OAuth and scope public.

        **Parameters**

        limit: Optional[:class:`int`]

        sort: Optional[Union[:class:`str`, :class:`MatchSort`]]

        **Returns**

        Dict[:class:`str`, Union[Dict, Sequence[:class:`Match`]]]
        """
        sort = parse_enum_args(sort)
        resp = await self.http.make_request('get', Path.get_matches(), limit=limit, sort=sort)
        return {
            "matches": list(map(Match, resp['matches'])),
            "cursor": resp['cursor'],
            "params": resp['params'],
        }

    async def get_match(self, match_id: int) -> Match:
        """
        Returns a match by id.

        Requires OAuth and scope public.

        **Parameters**

        match_id: :class:`int`
            The match id.

        **Returns**

        :class:`Match`
        """
        return MatchExtended(await self.http.make_request('get', Path.get_match(match_id)))

    async def get_rooms(self, mode: Union[str, GameModeStr] = '', sort: Union[str, RoomSort] = None) -> Sequence[Room]:
        """
        Returns a list of rooms.

        Requires OAuth and scope public.

        **Parameters**

        mode: Optional[Union[:class:`str`, :class:`GameModeStr`]]
            Game mode to filter rooms by.

        sort: Optional[Union[:class:`str`, :class:`RoomSort`]]
            Sort rooms by.
        """
        mode = parse_enum_args(mode)
        return list(map(Room, await self.http.make_request('get', Path.get_rooms(mode), sort=sort)))

    async def get_seasonal_backgrounds(self) -> SeasonalBackgrounds:
        """
        Get the season backgrounds.

        Doesn't require OAuth

        **Returns**

        :class:`SeasonalBackgrounds`
        """
        return SeasonalBackgrounds(await self.http.make_request('get', Path.get_seasonal_backgrounds()))

    async def get_room(self, room_id: int) -> Room:
        """
        Returns a room by id.

        Requires OAuth and scope public.

        **Parameters**

        room_id: :class:`int`
            The room id.

        **Returns**

        :class:`Room`
        """
        return Room(await self.http.make_request('get', Path.get_room(room_id)))

    async def get_score_by_id(self, mode, score_id) -> LegacyScore:
        """
        Returns a score by id.

        Requires OAuth and scope public.

        **Parameters**

        mode: Union[:class:`str`, :class:`GameModeStr`]

        score_id: :class:`int`

        **Returns**

        :class:`LegacyScore`
        """
        mode = parse_enum_args(mode)
        return LegacyScore(await self.http.make_request('get', Path.get_score_by_id(mode, score_id)))

    async def search_beatmapsets(self, filters=None, page=None) -> dict:
        """
        Search for beatmapsets.

        Requires OAuth and scope public.

        **Attributes**

        filters: Optional[:class:`BeatmapsetSearchFilter`]

        page: Optional[:class:`int`]

        **Returns**

        {

        "beatmapsets": Sequence[:class:`Beatmapset`]

        "cursor": :class:`dict`,

        "search": :class:`dict`,

        "recommended_difficulty": Union[:class:`float`, :class:`None`]

        "error": Union[:class:`str`, :class:`None`]

        "total": :class:`int`

        }
        """
        if filters is None:
            filters = {}
        if isinstance(filters, BeatmapsetSearchFilter):
            filters = filters.filters
        resp = await self.http.make_request('get', Path('beatmapsets/search', 'public'), page=page, **filters)
        return {
            'beatmapsets': [Beatmapset(beatmapset) for beatmapset in resp['beatmapsets']],
            'cursor': resp['cursor'],
            'search': resp['search'],
            'recommended_difficulty': resp['recommended_difficulty'],
            'error': resp['error'],
            'total': resp['total'],
        }

    async def get_room_leaderboard(self, room_id: int) -> \
            Dict[str, Union[Sequence[UserScoreAggregate], Union[UserScoreAggregate, None]]]:
        """
        Return a room's leaderboard. The :class:`UserScoreAggregate` objects returned under the "leaderboard"
        key contain the "user" attribute. The :class:`UserScoreAggregate` object under the "user_score" key
        contains the "user" and "position" attributes.

        Requires OAuth, scope public, and a user (authorization code grant or delegate scope).

        **Parameters**

        room_id: :class:`int`

        **Returns**

        Dict[:class:`str`, Union[Sequence[:class:`UserScoreAggregate`],
        Union[:class:`UserScoreAggregate`, :class:`NoneType`]]]

        {

            'leaderboard': Sequence[:class:`UserScoreAggregate`],

            'user_score': Union[:class:`UserScoreAggregate`, :class:`NoneType`]

        }
        """
        resp = await self.http.make_request('get', Path.get_room_leaderboard(room_id))
        return {
            'leaderboard': list(map(UserScoreAggregate, resp['leaderboard'])),
            'user_score': UserScoreAggregate(resp['user_score']) if resp['user_score'] is not None else None,
        }

    async def get_replay_data(self, mode, score_id):
        """
        Returns replay data for a score.

        Requires OAuth, scope public, and a user (authorization code grant or delegate scope).

        **Parameters**

        mode: Union[:class:`str`, :class:`GameModeStr`]

        score_id: :class:`int`

        **Returns**

        :class:`osrparse.Replay`
        """
        mode = parse_enum_args(mode)
        return Replay.from_string(await self.http.make_request('get', Path.get_replay_data(mode, score_id),
                                                               is_download=True))

    async def get_friends(self):
        """
        Returns a list of friends.

        Requires OAuth, scope friends.read, and a user (authorization code grant or delegate scope).

        **Returns**

        Sequence[:class:`User`]
        """
        return list(map(UserCompact, await self.http.make_request('get', Path.get_friends())))



from .beatmap import Beatmapset
from .user import UserStatistics
from ..util import prettify


class Rankings:
    """
    **Attributes**

    beatmapsets: Sequence[:Beatmapset:] or :class:`NoneType`
        The list of beatmaps in the requested spotlight for the given mode;
        only available if type is charts

    cursor: :class:`dict`
        To be used to query the next page

    ranking: Sequence[:class:`UserStatistics`]
        Score details ordered by rank in descending order.

    spotlight: :class:`Spotlight` or :class:`NoneType`
        Spotlight details; only available if type is charts

    total: :class:`int`
        An approximate count of ranks available
    """
    __slots__ = (
        "beatmapsets", "cursor", "ranking", "spotlight", "total"
    )

    def __init__(self, data):
        self.cursor = data['cursor']
        self.ranking = list(map(UserStatistics, data['ranking']))
        self.total = data['total']
        self.spotlight = Spotlight(data['spotlight']) if data.get('spotlight') is not None else None
        self.beatmapsets = list(map(Beatmapset, data['beatmapsets'])) if data.get('beatmapsets') is not None else None

    def __repr__(self):
        return prettify(self, 'ranking')


class Spotlight:
    """
    The details of a spotlight.

    **Attributes**

    end_date: :class:`str`
        In DateTime format. The end date of the spotlight.

    id: :class:`int`
        The ID of this spotlight.

    mode_specific: :class:`bool`
        If the spotlight has different mades specific to each :ref:`GameMode`.

    participant_count: :class:`int` or :class:`NoneType`
        The number of users participating in this spotlight. This is only shown when viewing a single spotlight.

    name: :class:`str`
        The name of the spotlight.

    start_date: :class:`str`
        In DatTime format. The starting date of the spotlight.

    type: :class:`str`
        The type of spotlight.
    """
    __slots__ = (
        "end_date", "id", "mode_specific", "name", "start_date", "type", "participant_count"
    )

    def __init__(self, data):
        self.end_date = data['end_date']
        self.id = data['id']
        self.mode_specific = data['mode_specific']
        self.participant_count = data['participant_count']
        self.name = data['name']
        self.start_date = data['start_date']
        self.type = data['type']

    def __repr__(self):
        return prettify(self, 'name')


class Spotlights:
    """
    **Attributes**

    spotlights: Sequence[:class:`Spotlight`]
    """
    __slots__ = (
        "spotlights",
    )

    def __init__(self, data):
        self.spotlights = list(map(Spotlight, data['spotlights']))

    def __repr__(self):
        return prettify(self, 'spotlights')



from ..util import prettify


class WikiPage:
    """
    Represents a wiki article

    **Attributes**

    available_locales: Sequence[:class:`str`]
        All available locales for the article.

    layout: :class:`str`
        The layout type for the page.

    locale: :class:`str`
        All lowercase BCP 47 language tag.

    markdown: :class:`str`
        Markdown content.

    path: :class:`str`
        Path of the article.

    subtitle: :class:`str` or :class:`NoneType`
        The article's subtitle.

    tags: Sequence[:class:`str`]
        Associated tags for the article.

    title: :class:`str`
        The article's title.
    """
    __slots__ = (
        'available_locales', 'layout', 'locale', 'markdown', 'path', 'subtitle',
        'tags', 'title'
    )

    def __init__(self, data):
        self.available_locales = data['available_locales']
        self.layout = data['layout']
        self.locale = data['locale']
        self.markdown = data['markdown']
        self.path = data['path']
        self.subtitle = data['subtitle']
        self.tags = data['tags']
        self.title = data['title']

    def __repr__(self):
        return prettify(self, 'title')



from ..enums import GameModeStr, BeatmapsetEventType
from .discussion import BeatmapsetDiscussion
from .beatmap import BeatmapsetCompact
from dateutil import parser

from ..util import prettify


class BeatmapsetEvent:
    """
    Represent a beatmapset event. This object is relevant for the :func:`osu.Client.get_beatmapset_events` endpoint.

    **Attributes**

    id: :class:`int`

    type: :class:`BeatmapsetEventType`

    comment: Union[:class:`BeatmapsetEventComment`, :class:`NoneType`]
        Is :class:`NoneType` for the following types:
        - :class:`BeatmapsetEventType.LOVE`
        - :class:`BeatmapsetEventType.QUALIFY`
        - :class:`BeatmapsetEventType.APPROVE`
        - :class:`BeatmapsetEventType.RANK`

    created_at: :class:`datetime.datetime`

    user_id: Union[:class:`int`, :class:`NoneType`]

    beatmapset: Union[:class:`BeatmapsetCompact`, :class:`NoneType`]

    discussion: Union[:class:`BeatmapsetDiscussion`, :class:`NoneType`]
    """
    __slots__ = (
        "id", "type", "comment", "created_at", "user_id", "beatmapset", "discussion"
    )

    def __init__(self, data):
        self.id = data['id']
        self.type = BeatmapsetEventType(data['type'])
        self.comment = data['comment']
        if self.comment is not None:
            self.comment = BeatmapsetEventComment(self.comment, self.type)
        self.created_at = parser.parse(data['created_at'])
        self.user_id = data.get('user_id')
        self.beatmapset = BeatmapsetCompact(data['beatmapset']) if data.get("beatmapset") is not None else None
        self.discussion = BeatmapsetDiscussion(data['discussion']) if data.get('discussion') is not None else None

    def __repr__(self):
        return prettify(self, 'type', 'beatmapset')


class BeatmapsetEventComment:
    """
    This object holds some extra information of the event.

    **Attributes**

    beatmap_discussion_id: Union[:class:`int`, :class:`NoneType`]

    beatmap_discussion_post_id: Union[:class:`int`, :class:`NoneType`]

    event_data: Union[:class:`BeatmapsetEventNominate`, :class:`BeatmapsetEventRemoveFromLoved`,
    :class:`BeatmapsetEventDisqualify`, :class:`BeatmapsetEventKudosuGain`, :class:`BeatmapsetEventKudosuLost`,
    :class:`BeatmapsetEventKudosuRecalculate`, :class:`BeatmapsetEventDiscussionLock`,
    :class:`BeatmapsetEventNominationReset`, :class:`BeatmapsetEventNominationResetReceived`,
    :class:`BeatmapsetEventGenreEdit`, :class:`BeatmapsetEventLanguageEdit`, :class:`BeatmapsetEventNsfwToggle`,
    :class:`BeatmapsetEventOffsetEdit`, :class:`BeatmapsetEventBeatmapOwnerChange`]
        The type of this attribute depends on the type of the BeatmapsetEvent object.
    """

    __slots__ = (
        "beatmap_discussion_id", "beatmap_discussion_post_id", "event_data"
    )
    event_types = {
        BeatmapsetEventType.NOMINATE: 'BeatmapsetEventNominate',
        BeatmapsetEventType.REMOVE_FROM_LOVED: 'BeatmapsetEventRemoveFromLoved',
        BeatmapsetEventType.DISQUALIFY: 'BeatmapsetEventDisqualify',
        BeatmapsetEventType.KUDOSU_GAIN: 'BeatmapsetEventKudosuGain',
        BeatmapsetEventType.KUDOSU_LOST: 'BeatmapsetEventKudosuLost',
        BeatmapsetEventType.KUDOSU_RECALCULATE: 'BeatmapsetEventKudosuRecalculate',
        BeatmapsetEventType.DISCUSSION_LOCK: 'BeatmapsetEventDiscussionLock',
        BeatmapsetEventType.NOMINATION_RESET: 'BeatmapsetEventNominationReset',
        BeatmapsetEventType.NOMINATION_RESET_RECEIVED: 'BeatmapsetEventNominationResetReceived',
        BeatmapsetEventType.GENRE_EDIT: 'BeatmapsetEventGenreEdit',
        BeatmapsetEventType.LANGUAGE_EDIT: 'BeatmapsetEventLanguageEdit',
        BeatmapsetEventType.NSFW_TOGGLE: 'BeatmapsetEventNsfwToggle',
        BeatmapsetEventType.OFFSET_EDIT: 'BeatmapsetEventOffsetEdit',
        BeatmapsetEventType.BEATMAP_OWNER_CHANGE: 'BeatmapsetEventBeatmapOwnerChange',
    }

    def __init__(self, data, type):
        self.beatmap_discussion_id = data.get('beatmap_discussion_id')
        self.beatmap_discussion_post_id = data.get('beatmap_discussion_post_id')
        self.event_data = globals()[self.event_types[type]](data) if type in self.event_types else None

    def __repr__(self):
        return prettify(self, 'beatmap_discussion_id', 'beatmap_discussion_post_id')


class BeatmapsetEventNominate:
    """
    **Attributes**

    modes: Sequence[:class:`GameModeStr`]
    """
    __slots__ = ("modes",)

    def __init__(self, data):
        self.modes = [GameModeStr(mode) for mode in data['modes']]

    def __repr__(self):
        return prettify(self, 'modes')


class BeatmapsetEventRemoveFromLoved:
    """
    **Attributes**

    reason: :class:`str`
    """
    __slots__ = ("reason",)

    def __init__(self, data):
        self.reason = data['reason']

    def __repr__(self):
        return prettify(self, 'reason')


class BeatmapsetEventDisqualify:
    """
    **Attributes**

    nominator_ids: Sequence[:class:`int`]
    """
    __slots__ = ("nominator_ids",)

    def __init__(self, data):
        self.nominator_ids = data['nominator_ids']

    def __repr__(self):
        return prettify(self, 'nominator_ids')


class BeatmapsetEventVote:
    """
    **Attributes**

    user_id: :class:`int`

    score: :class:`int`
    """
    __slots__ = ("user_id", "score")

    def __init__(self, data):
        self.user_id = data['user_id']
        self.score = data['score']

    def __repr__(self):
        return prettify(self, 'user_id', 'score')


class BeatmapsetEventKudosuChange:
    """
    **Attributes**

    new_votes: Union[:class:`BeatmapsetEventVote`, :class:`NoneType`]

    votes: Union[Sequence[:class:`BeatmapsetEventVote`], :class:`NoneType`]
    """
    __slots__ = ("new_votes", "votes")

    def __init__(self, data):
        self.new_votes = BeatmapsetEventVote(data['new_votes']) if data.get('new_votes') is not None else None
        self.votes = list(map(BeatmapsetEventVote, data['votes'])) if data.get('votes') is not None else None

    def __repr__(self):
        return prettify(self, 'new_votes', 'votes')


class BeatmapsetEventKudosuGain(BeatmapsetEventKudosuChange):
    __doc__ = BeatmapsetEventKudosuChange.__doc__


class BeatmapsetEventKudosuLost(BeatmapsetEventKudosuChange):
    __doc__ = BeatmapsetEventKudosuChange.__doc__


class BeatmapsetEventKudosuRecalculate:
    """
    **Attributes**

    new_votes: Union[:class:`BeatmapsetEventVote`, :class:`NoneType`]
    """
    __slots__ = ("new_votes",)

    def __init__(self, data):
        self.new_votes = BeatmapsetEventVote(data['new_votes']) if data.get('new_votes') is not None else None

    def __repr__(self):
        return prettify(self, 'new_votes')


class BeatmapsetEventDiscussionLock:
    """
    **Attributes**

    reason: :class:`str`
    """
    __slots__ = ("reason",)

    def __init__(self, data):
        self.reason = data['reason']

    def __repr__(self):
        return prettify(self, 'reason')


class BeatmapsetEventNominationReset:
    """
    **Attributes**

    nominator_ids: Sequence[:class:`int`]
    """
    __slots__ = ("nominator_ids",)

    def __init__(self, data):
        self.nominator_ids = data['nominator_ids']

    def __repr__(self):
        return prettify(self, 'nominator_ids')


class BeatmapsetEventNominationResetReceived:
    """
    **Attributes**

    source_user_id: :class:`int`

    source_user_name: :class:`str`
    """
    __slots__ = ("source_user_id", "source_user_username")

    def __init__(self, data):
        self.source_user_id = data['source_user_id']
        self.source_user_username = data['source_user_username']

    def __repr__(self):
        return prettify(self, 'source_user_id', 'source_user_username')


class BeatmapsetEventEdit:
    """
    **Attributes**

    old: :class:`str`

    new: :class:`str`
    """
    __slots__ = ("old", "new")

    def __init__(self, data):
        self.old = data['old']
        self.new = data['new']

    def __repr__(self):
        return prettify(self, 'old', 'new')


class BeatmapsetEventGenreEdit(BeatmapsetEventEdit):
    __doc__ = BeatmapsetEventEdit.__doc__


class BeatmapsetEventLanguageEdit(BeatmapsetEventEdit):
    __doc__ = BeatmapsetEventEdit.__doc__


class BeatmapsetEventNsfwToggle(BeatmapsetEventEdit):
    """
    **Attributes**

    old: :class:`bool`

    new: :class:`bool`
    """


class BeatmapsetEventOffsetEdit(BeatmapsetEventEdit):
    """
    **Attributes**

    old: :class:`int`

    new: :class:`int`
    """


class BeatmapsetEventBeatmapOwnerChange:
    """
    **Attributes**

    beatmap_id: :class:`int`

    beatmap_version: :class:`str`

    new_user_id: :class:`int`

    new_user_username: :class:`str`
    """
    __slots__ = ("beatmap_id", "beatmap_version", "new_user_id", "new_user_username")

    def __init__(self, data):
        self.beatmap_id = data['beatmap_id']
        self.beatmap_version = data['beatmap_version']
        self.new_user_id = data['new_user_id']
        self.new_user_username = data['new_user_username']

    def __repr__(self):
        return prettify(self, 'beatmap_id', 'new_user_username')



import websockets
import asyncio
import json

from .objects import Notification


class NotificationWebsocket:
    """
    This class allows you to receive notifications without constantly polling the server.
    To utilize it you should do either of:
        - Make a class inheriting this one and redefine the event functions (on_logout, on_new, ...).
        - Use the event function as a decorator, read more on its use under its docs.

    **Event types**

    on_logout
        Server will disconnect session after sending this event so don't try to reconnect.

    on_new
        New notification. See :class:`Notification` object for notification types.

        **Arguments**

        notification: :class:`Notification`

    on_read
        Notification has been read.

        **Arguments**

        ids: :class:`int`
            list of ids of Notifications which are read.

    on_unplanned_disconnect
        Event fired by NotificationWebsocket object when connection without having been sent a logout event.
        Default function fires connect to try and reconnect.
    """
    valid_events = [
        'logout', 'new', 'read', 'unplanned_disconnect'
    ]

    def __init__(self, notification_uri, auth):
        """
        **Arguments**

        notification_url: :class:`str`
            url endpoint to connect to. Can obtain one via get_notifications

        auth: :class:`AuthHandler`
            The same auth handler used for Client
        """

        self.auth = auth
        self.uri = notification_uri
        self.loop = asyncio.get_event_loop()
        self.ws = None
        self.connected = False

    async def _run(self):
        headers = {
            "Authorization": f"Bearer {self.auth.token}"
        }
        async with websockets.connect(self.uri, headers=headers) as ws:
            self.ws = ws
            self.connected = True
            while self.connected:
                try:
                    event = await self.ws.recv()
                except websockets.exceptions.ConnectionClosed:
                    self.ws = None
                    self.connected = False
                    self._on_unplanned_disconnect()
                    return

                event = json.loads(event)
                event_type = event['event']
                del event['event']

                getattr(self, "_"+event_type, event_type)(*list(event.values()))

    def connect(self):
        """
        Connect to the uri and start receiving events.
        Runs with asyncio.run_coroutine_threadsafe

        **Returns**

        :class:`concurrent.futures.Future`
        """
        return asyncio.run_coroutine_threadsafe(self._run())

    def event(self, func):
        """
        Decorator for adding event functions. Example:

        .. code-block:: Python

            notification_websocket = NotificationWebsocket(notif_uri, auth)

            @notification_websocket.event()
            def new(notification):
                print(notification.name)
        """
        if func.__name__ not in self.valid_events:
            raise NameError(f"This is not a valid event name. Valid events consist of {', '.join(self.valid_events)}")
        setattr(self, func.__name__, func)

    # Default events

    def _on_logout(self):
        self.connected = False
        if hasattr(self, 'on_logout'):
            self.on_logout()

    def _on_new(self, data):
        data = Notification(data)
        if hasattr(self, 'on_new'):
            self.on_new(data)

    def _on_read(self, ids):
        if hasattr(self, 'on_read'):
            self.on_read(ids)

    def _on_unplanned_disconnect(self):
        # Event fired by NotificationWebsocket object when connection without having been sent a logout event.
        # Default function fires connect to try and reconnect.
        self.connect()
        if hasattr(self, 'on_unplanned_disconnect'):
            self.on_unplanned_disconnect()



from .user import UserCompact
from .beatmap import BeatmapCompact
from .score import ScoreDataStatistic, LazerMod
from ..enums import RoomCategory, RoomType, RealTimeQueueMode, PlaylistQueueMode, GameModeInt
from ..util import prettify
from dateutil import parser


class MultiplayerScore:
    """
    Score data.

    **Attributes**

    id: :class:`int`

    user_id: :class:`int`

    room_id: :class:`int`

    playlist_item_id: :class:`int`

    beatmap_id: :class:`int`

    rank: :class:`str`
        Can be one of the following: charts (Spotlight), country (Country), performance (Performance), score (Score)

    total_score: :class:`int`

    accuracy: :class:`int`

    max_combo: :class:`int`

    mods: Sequence[:class:`LazerMod`]

    statistics: :class:`ScoreStatistics`

    passed: :class:`bool`

    position: :class:`int` or :class:`NoneType`

    scores_around: :class:`MultiplayerScoresAround` or :class:`NoneType`
        Scores around the specified score.

    user: :class:`User`
    """
    __slots__ = (
        "id", "user_id", "room_id", "playlist_item_id", "beatmap_id", "rank",
        "total_score", "accuracy", "max_combo", "mods", "statistics", "passed",
        "position", "scores_around", "user"
    )

    def __init__(self, data):
        self.id = data['id']
        self.user_id = data['user_id']
        self.room_id = data['room_id']
        self.playlist_item_id = data['playlist_item_id']
        self.beatmap_id = data['beatmap_id']
        self.rank = data['rank']
        self.total_score = data['total_score']
        self.accuracy = data['accuracy']
        self.max_combo = data['max_combo']
        self.mods = list(map(LazerMod, data['mods'])) if data["mods"] is not None else []
        self.statistics = ScoreDataStatistic(data['statistics'])
        self.passed = data['passed']
        self.position = data.get('position')
        self.scores_around = MultiplayerScoresAround(data['scores_around']) \
            if data.get('scores_around') is not None else None
        self.user = UserCompact(data['user']) if data.get('user') is not None else None

    def __repr__(self):
        return prettify(self, 'user', 'position')


class MultiplayerScores:
    """
    An object which contains scores and related data for fetching next page of the result.
    To fetch the next page, make request to scores index (Client.get_scores) with relevant
    room and playlist, use the data in attribute params and cursor to fill in the 3 other optional queries.

    **Attributes**

    cursor: :class:`dict`
        To be used to fetch the next page.

    params: :class:`dict`
        To be used to fetch the next page.

    scores: Sequence[:class:`MultiplayerScore`]

    total: :class:`int` or :class:`NoneType`
        Index only. Total scores of the specified playlist item.

    user_score: :class:`MultiplayerScore` or :class:`NoneType`
        Index only. Score of the accessing user if exists.
    """
    __slots__ = (
        "cursor", "params", "scores", "total", "user_score"
    )

    def __init__(self, data):
        self.cursor = data['cursor']
        self.params = data['params']
        self.scores = list(map(MultiplayerScore, data['scores'])) if data['scores'] is not None else None
        self.total = data['total']
        self.user_score = MultiplayerScore(data['user_score']) if data['user_score'] is not None else None

    def __repr__(self):
        return prettify(self, 'total', 'scores')


class MultiplayerScoresAround:
    """
    **Attributes**

    higher: :class:`MultiplayerScores`

    lower: :class:`MultiplayerScores`
    """
    __slots__ = (
        "higher", "lower"
    )

    def __init__(self, data):
        self.higher = MultiplayerScores(data['higher'])
        self.lower = MultiplayerScores(data['lower'])

    def __repr__(self):
        return prettify(self, 'higher', 'lower')


class Room:
    """
    :func:`osu.Client.get_rooms` and :func:`osu.Client.get_room` endpoints include host, playlist,
    and recent_participants attributes. In addition, the :class:`PlaylistItem` objects in playlist
    include the beatmap attribute and the beatmap attribute has the beatmapset, checksum, and
    max_combo attributes.

    **Attributes**

    id: :class:`int`

    name: :class:`str`

    category: :class:`RoomCategory`

    type: :class:`RoomType`

    user_id: :class:`int`

    start_at: Union[:class:`datetime.datetime`, :class:`NoneType`]

    ends_at: Union[:class:`datetime.datetime`, :class:`NoneType`]

    max_attempts: :class:`int`

    participant_count: :class:`int`

    channel_id: :class:`int`

    active: :class:`bool`

    has_password: :class:`bool`

    queue_mode: Union[:class:`RealTimeQueueMode`, :class:`PlaylistQueueMode`]
        type depends on the room type. :class:`RealTimeQueueMode` for type :class:`RoomType`.REALTIME and
        :class:`PlaylistQueueMode` for type :class:`RoomType`.PLAYLIST.

    current_playlist_item: Union[:class:`PlaylistItem`, :class:`NoneType`]

    current_user_score: Union[:class:`UserScoreAggregate`, :class:`NoneType`]

    difficulty_range: Union[Dict[:class:`str`, :class:`int`], :class:`NoneType`]
        When not none, is a dictionary containing keys "max" and "min."

    host: Union[:class:`UserCompact`, :class:`NoneType`]

    playlist: Union[Sequence[:class:`PlaylistItem`], :class:`NoneType`]

    playlist_item_stats: Union[:class:`PlaylistItemStats`, :class:`NoneType`]

    recent_participants: Union[Sequence[:class:`UserCompact`], :class:`NoneType`]

    scores: Union[:class:`MultiplayerScore`, :class:`NoneType`]
    """
    __slots__ = (
        "id", "name", "category", "type", "user_id", "start_at", "ends_at", "max_attempts",
        "participant_count", "channel_id", "active", "has_password", "queue_mode",
        "current_playlist_item", "current_user_score", "difficulty_range", "host",
        "playlist", "playlist_item_stats", "recent_participants", "scores"
    )

    def __init__(self, data):
        self.id = data['id']
        self.name = data['name']
        self.category = RoomCategory(data['category'])
        self.type = RoomType(data['type'])
        self.user_id = data['user_id']
        self.start_at = parser.parse(data['start_at']) if data.get('start_at') is not None else None
        self.ends_at = parser.parse(data['ends_at']) if data.get('ends_at') is not None else None
        self.max_attempts = data['max_attempts']
        self.participant_count = data['participant_count']
        self.channel_id = data['channel_id']
        self.active = data['active']
        self.has_password = data['has_password']
        self.queue_mode = (PlaylistQueueMode if self.type == RoomType.PLAYLISTS
                           else RealTimeQueueMode)(data['queue_mode'])

        self.current_playlist_item = PlaylistItem(data['current_playlist_item']) \
            if data.get('current_playlist_item') is not None else None
        self.current_user_score = UserScoreAggregate(data['current_user_score']) \
            if data.get('current_user_score') is not None else None
        self.difficulty_range = data.get('difficulty_range')  # {min, max}
        self.host = UserCompact(data['host']) if data.get('host') is not None else None
        self.playlist = list(map(PlaylistItem, data['playlist'])) if data.get('playlist') is not None else None
        self.playlist_item_stats = PlaylistItemStats(data['playlist_item_stats']) \
            if data.get('playlist_item_stats') is not None else None
        self.recent_participants = list(map(UserCompact, data['recent_participants'])) \
            if data.get('recent_participants') is not None else None
        self.scores = list(map(MultiplayerScore, data['scores'])) if data.get('scores') is not None else None

    def __repr__(self):
        return prettify(self, 'name', 'type')


class UserScoreAggregate:
    """
    **Attributes**

    accuracy: :class:`float`

    attempts: :class:`int`

    completed: :class:`int`

    pp: :class:`float`

    room_id: :class:`int`

    total_score: :class:`int`

    user_id: :class:`int`

    playlist_item_attempts: Union[Dict[:class:`str`, :class:`int`], :class:`NoneType`]
        contains keys "attempts" and "id."

    position: Union[:class:`int`, :class:`NoneType`]
        user rank

    user: Union[:class:`UserCompact`, :class:`NoneType`]
    """
    __slots__ = (
        "accuracy", "attempts", "completed", "pp", "room_id", "total_score",
        "user_id", "playlist_item_attempts", "position", "user"
    )

    def __init__(self, data):
        self.accuracy = data['accuracy']
        self.attempts = data['attempts']
        self.completed = data['completed']
        self.pp = data['pp']
        self.room_id = data['room_id']
        self.total_score = data['total_score']
        self.user_id = data['user_id']
        self.playlist_item_attempts = data.get('playlist_item_attempts')  # {attempts, id}
        self.position = data.get('position')
        self.user = UserCompact(data['user']) if data.get('user') is not None else None

    def __repr__(self):
        attributes = ('total_score', 'user' if self.user is not None else 'user_id')
        return prettify(self, *attributes)


class PlaylistItemStats:
    """
    **Attributes**

    count_active: :class:`int`

    count_total: :class:`int`

    ruleset_ids: :class:`Sequence[:class:`GameModeInt`]
    """
    __slots__ = ("count_active", "count_total", "ruleset_ids")

    def __init__(self, data):
        self.count_active = data['count_active']
        self.count_total = data['count_total']
        self.ruleset_ids = list(map(GameModeInt, data['ruleset_ids']))

    def __repr__(self):
        return prettify(self, 'count_active', 'count_total')


class PlaylistItem:
    """
    **Attributes**

    id: :class:`int`

    room_id: :class:`int`

    beatmap_id: :class:`int`

    ruleset_id: :class:`GameModeInt`

    allowed_mods: Sequence[:class:`LazerMod`]

    required_mods: Sequence[:class:`LazerMod`]

    expired: :class:`bool`

    owner_id: :class:`int`

    playlist_order: Union[:class:`int`, :class:`NoneType`]

    played_at: Union[:class:`datetime.datetime`, :class:`NoneType`]

    beatmap: Union[:class:`BeatmapCompact`, :class:`NoneType`]
    """
    __slots__ = (
        "id", "room_id", "beatmap_id", "ruleset_id", "allowed_mods", "required_mods",
        "expired", "owner_id", "playlist_order", "played_at", "beatmap"
    )

    def __init__(self, data):
        self.id = data['id']
        self.room_id = data['room_id']
        self.beatmap_id = data['beatmap_id']
        self.ruleset_id = GameModeInt(data['ruleset_id'])
        self.allowed_mods = list(map(LazerMod, data['allowed_mods']))
        self.required_mods = list(map(LazerMod, data['required_mods']))
        self.expired = data['expired']
        self.owner_id = data['owner_id']
        self.playlist_order = data['playlist_order']
        self.played_at = parser.parse(data['played_at']) if data['played_at'] is not None else None
        self.beatmap = BeatmapCompact(data['beatmap']) if data.get('beatmap') is not None else None

    def __repr__(self):
        attributes = ('id', 'beatmap' if self.beatmap is not None else 'beatmap_id')
        return prettify(self, *attributes)



class Scope:
    """
    Scope object for telling the program what scopes you are using

    **Valid scopes**

    public
        Allows reading of publicly available data on behalf of the user.
    identify (default)
        Allows reading of the public profile of the user (/me).
    friends.read
        Allows reading of the user's friend list.
    forum.write
        Allows creating and editing forum posts on a user's behalf.
    delegate
        Allows acting as the owner of a client; only available for Client Credentials Grant.
    chat.write
        Allows sending chat messages on a user's behalf.
    """
    valid_scopes = [
        'chat.write',
        'delegate',
        'forum.write',
        'friends.read',
        'identify',
        'public',
    ]

    def __init__(self, *scopes):
        for scope in scopes:
            if scope not in self.valid_scopes:
                raise NameError(f"{scope} is not a valid scope. The valid scopes consist of "
                                f"{','.join(self.valid_scopes)}")
        self.scopes = ' '.join(scopes)
        self.scopes_list = list(scopes)

    @classmethod
    def default(cls):
        return cls('public')

    @classmethod
    def identify(cls):
        return cls('public', 'identify')

    def __iter__(self):
        return iter(self.scopes_list)

    def __len__(self):
        return len(self.scopes_list)

    def __str__(self):
        return str(self.scopes_list)

    def __contains__(self, scope: str):
        return scope in self.scopes_list
